{
 "cells": [
  {
   "cell_type": "markdown",
   "id": "7776b2ee-ffa6-49a7-a046-e6e465a23888",
   "metadata": {},
   "source": [
    "# Data Grouping and Aggregation with Kotlin"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 1,
   "id": "2a0e91d4-11cc-4aeb-ba14-e7d7a4ca9b51",
   "metadata": {
    "tags": []
   },
   "outputs": [
    {
     "data": {
      "text/html": [
       "            <div id=\"kotlin_out_0\"/>\n",
       "            <script type=\"text/javascript\">\n",
       "                            if(!window.kotlinQueues) {\n",
       "                window.kotlinQueues = {};\n",
       "            }\n",
       "            if(!window.kotlinQueues[\"DataFrame\"]) {\n",
       "                var resQueue = [];\n",
       "                window.kotlinQueues[\"DataFrame\"] = resQueue;\n",
       "                window[\"call_DataFrame\"] = function(f) {\n",
       "                    resQueue.push(f);\n",
       "                }\n",
       "            }\n",
       "            (function (){\n",
       "                var modifiers = [(function(script) {\n",
       "    script.src = \"https://cdn.jsdelivr.net/gh/Kotlin/dataframe@3db46ccccaa1291c0627307d64133317f545e6ae/core/src/main/resources/init.js\"\n",
       "    script.type = \"text/javascript\";\n",
       "})];\n",
       "                var e = document.getElementById(\"kotlin_out_0\");\n",
       "                modifiers.forEach(function (gen) {\n",
       "                    var script = document.createElement(\"script\");\n",
       "                    gen(script)\n",
       "                    script.addEventListener(\"load\", function() {\n",
       "                        window[\"call_DataFrame\"] = function(f) {f();};\n",
       "                        window.kotlinQueues[\"DataFrame\"].forEach(function(f) {f();});\n",
       "                        window.kotlinQueues[\"DataFrame\"] = [];\n",
       "                    }, false);\n",
       "                    script.addEventListener(\"error\", function() {\n",
       "                        window[\"call_DataFrame\"] = function(f) {};\n",
       "                        window.kotlinQueues[\"DataFrame\"] = [];\n",
       "                        var div = document.createElement(\"div\");\n",
       "                        div.style.color = 'darkred';\n",
       "                        div.textContent = 'Error loading resource DataFrame';\n",
       "                        document.getElementById(\"kotlin_out_0\").appendChild(div);\n",
       "                    }, false);\n",
       "                    \n",
       "                    e.appendChild(script);\n",
       "                });\n",
       "            })();\n",
       "            </script>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    },
    {
     "data": {
      "text/html": [
       "                <style>\n",
       "                :root {\n",
       "    --background: #fff;\n",
       "    --background-odd: #f5f5f5;\n",
       "    --background-hover: #d9edfd;\n",
       "    --header-text-color: #474747;\n",
       "    --text-color: #848484;\n",
       "    --text-color-dark: #000;\n",
       "    --text-color-medium: #737373;\n",
       "    --text-color-pale: #b3b3b3;\n",
       "    --inner-border-color: #aaa;\n",
       "    --bold-border-color: #000;\n",
       "    --link-color: #296eaa;\n",
       "    --link-color-pale: #296eaa;\n",
       "    --link-hover: #1a466c;\n",
       "}\n",
       "\n",
       ":root[theme=\"dark\"], :root [data-jp-theme-light=\"false\"], .dataframe_dark{\n",
       "    --background: #303030;\n",
       "    --background-odd: #3c3c3c;\n",
       "    --background-hover: #464646;\n",
       "    --header-text-color: #dddddd;\n",
       "    --text-color: #b3b3b3;\n",
       "    --text-color-dark: #dddddd;\n",
       "    --text-color-medium: #b2b2b2;\n",
       "    --text-color-pale: #737373;\n",
       "    --inner-border-color: #707070;\n",
       "    --bold-border-color: #777777;\n",
       "    --link-color: #008dc0;\n",
       "    --link-color-pale: #97e1fb;\n",
       "    --link-hover: #00688e;\n",
       "}\n",
       "\n",
       "p.dataframe_description {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "table.dataframe {\n",
       "    font-family: \"Helvetica Neue\", Helvetica, Arial, sans-serif;\n",
       "    font-size: 12px;\n",
       "    background-color: var(--background);\n",
       "    color: var(--text-color-dark);\n",
       "    border: none;\n",
       "    border-collapse: collapse;\n",
       "}\n",
       "\n",
       "table.dataframe th, td {\n",
       "    padding: 6px;\n",
       "    border: 1px solid transparent;\n",
       "    text-align: left;\n",
       "}\n",
       "\n",
       "table.dataframe th {\n",
       "    background-color: var(--background);\n",
       "    color: var(--header-text-color);\n",
       "}\n",
       "\n",
       "table.dataframe td {\n",
       "    vertical-align: top;\n",
       "}\n",
       "\n",
       "table.dataframe th.bottomBorder {\n",
       "    border-bottom-color: var(--bold-border-color);\n",
       "}\n",
       "\n",
       "table.dataframe tbody > tr:nth-child(odd) {\n",
       "    background: var(--background-odd);\n",
       "}\n",
       "\n",
       "table.dataframe tbody > tr:nth-child(even) {\n",
       "    background: var(--background);\n",
       "}\n",
       "\n",
       "table.dataframe tbody > tr:hover {\n",
       "    background: var(--background-hover);\n",
       "}\n",
       "\n",
       "table.dataframe a {\n",
       "    cursor: pointer;\n",
       "    color: var(--link-color);\n",
       "    text-decoration: none;\n",
       "}\n",
       "\n",
       "table.dataframe tr:hover > td a {\n",
       "    color: var(--link-color-pale);\n",
       "}\n",
       "\n",
       "table.dataframe a:hover {\n",
       "    color: var(--link-hover);\n",
       "    text-decoration: underline;\n",
       "}\n",
       "\n",
       "table.dataframe img {\n",
       "    max-width: fit-content;\n",
       "}\n",
       "\n",
       "table.dataframe th.complex {\n",
       "    background-color: var(--background);\n",
       "    border: 1px solid var(--background);\n",
       "}\n",
       "\n",
       "table.dataframe .leftBorder {\n",
       "    border-left-color: var(--inner-border-color);\n",
       "}\n",
       "\n",
       "table.dataframe .rightBorder {\n",
       "    border-right-color: var(--inner-border-color);\n",
       "}\n",
       "\n",
       "table.dataframe .rightAlign {\n",
       "    text-align: right;\n",
       "}\n",
       "\n",
       "table.dataframe .expanderSvg {\n",
       "    width: 8px;\n",
       "    height: 8px;\n",
       "    margin-right: 3px;\n",
       "}\n",
       "\n",
       "table.dataframe .expander {\n",
       "    display: flex;\n",
       "    align-items: center;\n",
       "}\n",
       "\n",
       "/* formatting */\n",
       "\n",
       "table.dataframe .null {\n",
       "    color: var(--text-color-pale);\n",
       "}\n",
       "\n",
       "table.dataframe .structural {\n",
       "    color: var(--text-color-medium);\n",
       "    font-weight: bold;\n",
       "}\n",
       "\n",
       "table.dataframe .dataFrameCaption {\n",
       "    font-weight: bold;\n",
       "}\n",
       "\n",
       "table.dataframe .numbers {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "table.dataframe td:hover .formatted .structural, .null {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "table.dataframe tr:hover .formatted .structural, .null {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "\n",
       "                </style>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    }
   ],
   "source": [
    "%use dataframe"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "62ede64f-a82d-4b29-b90b-a6a0e1ea402b",
   "metadata": {},
   "source": [
    "Kotlin DataFrame provides a very similar interface to the one supplied by pandas. The methods that we will use more will be `groupBy()` and `pivot()`, that we have already discussed in a previous chapter."
   ]
  },
  {
   "cell_type": "markdown",
   "id": "de6ad58c-de08-475c-b674-9ebbc06c04e0",
   "metadata": {},
   "source": [
    "Having no such thing as `pandas.Index`, with Kotlin DataFrame the structure of the tabular data is more flexible, but in the same time it lacks all the power that a *labeled* index or multi index could give.\n",
    "The tabular object represented by a Kotlin `DataFrame` is slightly different from the pandas one, and so we have to think differently sometimes when it comes to replicate some pandas behaviors, especially when pivoting and grouping by a variable."
   ]
  },
  {
   "cell_type": "markdown",
   "id": "e9b55684-5aa0-40b0-80e0-6721e7717eee",
   "metadata": {
    "tags": []
   },
   "source": [
    "## Split"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "4536f73a-a2cb-4d09-91b5-2942e00d1531",
   "metadata": {},
   "source": [
    "The \"Split\" operation in both pandas and DataFrame, is performed when invoking the `groupBy` or `pivot` operation. The result of those methods are not a dataset, but a temporary data structure that needs an aggregate operation to combine all the groups that are been created.\n",
    "\n",
    "Let's consider the following dataset:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 2,
   "id": "56ce4dac-a8f7-442e-b21e-369767e65e44",
   "metadata": {
    "tags": []
   },
   "outputs": [
    {
     "data": {
      "application/kotlindataframe+json": "{\"nrow\":10,\"ncol\":2,\"columns\":[\"letter\",\"num\"],\"kotlin_dataframe\":[{\"letter\":\"B\",\"num\":0.6246292191371761},{\"letter\":\"A\",\"num\":-0.8581918080882499},{\"letter\":\"B\",\"num\":0.6762208162903859},{\"letter\":\"C\",\"num\":1.126393826325953},{\"letter\":\"C\",\"num\":1.4376621993807677},{\"letter\":\"B\",\"num\":1.4920787730184157},{\"letter\":\"C\",\"num\":-1.5985374620272976},{\"letter\":\"B\",\"num\":-0.07756122563035872},{\"letter\":\"A\",\"num\":1.9917663922907851},{\"letter\":\"A\",\"num\":0.27096900216864345}]}",
      "text/html": [
       "        <html>\n",
       "        <head>\n",
       "            <style type=\"text/css\">\n",
       "                :root {\n",
       "    --background: #fff;\n",
       "    --background-odd: #f5f5f5;\n",
       "    --background-hover: #d9edfd;\n",
       "    --header-text-color: #474747;\n",
       "    --text-color: #848484;\n",
       "    --text-color-dark: #000;\n",
       "    --text-color-medium: #737373;\n",
       "    --text-color-pale: #b3b3b3;\n",
       "    --inner-border-color: #aaa;\n",
       "    --bold-border-color: #000;\n",
       "    --link-color: #296eaa;\n",
       "    --link-color-pale: #296eaa;\n",
       "    --link-hover: #1a466c;\n",
       "}\n",
       "\n",
       ":root[theme=\"dark\"], :root [data-jp-theme-light=\"false\"], .dataframe_dark{\n",
       "    --background: #303030;\n",
       "    --background-odd: #3c3c3c;\n",
       "    --background-hover: #464646;\n",
       "    --header-text-color: #dddddd;\n",
       "    --text-color: #b3b3b3;\n",
       "    --text-color-dark: #dddddd;\n",
       "    --text-color-medium: #b2b2b2;\n",
       "    --text-color-pale: #737373;\n",
       "    --inner-border-color: #707070;\n",
       "    --bold-border-color: #777777;\n",
       "    --link-color: #008dc0;\n",
       "    --link-color-pale: #97e1fb;\n",
       "    --link-hover: #00688e;\n",
       "}\n",
       "\n",
       "p.dataframe_description {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "table.dataframe {\n",
       "    font-family: \"Helvetica Neue\", Helvetica, Arial, sans-serif;\n",
       "    font-size: 12px;\n",
       "    background-color: var(--background);\n",
       "    color: var(--text-color-dark);\n",
       "    border: none;\n",
       "    border-collapse: collapse;\n",
       "}\n",
       "\n",
       "table.dataframe th, td {\n",
       "    padding: 6px;\n",
       "    border: 1px solid transparent;\n",
       "    text-align: left;\n",
       "}\n",
       "\n",
       "table.dataframe th {\n",
       "    background-color: var(--background);\n",
       "    color: var(--header-text-color);\n",
       "}\n",
       "\n",
       "table.dataframe td {\n",
       "    vertical-align: top;\n",
       "}\n",
       "\n",
       "table.dataframe th.bottomBorder {\n",
       "    border-bottom-color: var(--bold-border-color);\n",
       "}\n",
       "\n",
       "table.dataframe tbody > tr:nth-child(odd) {\n",
       "    background: var(--background-odd);\n",
       "}\n",
       "\n",
       "table.dataframe tbody > tr:nth-child(even) {\n",
       "    background: var(--background);\n",
       "}\n",
       "\n",
       "table.dataframe tbody > tr:hover {\n",
       "    background: var(--background-hover);\n",
       "}\n",
       "\n",
       "table.dataframe a {\n",
       "    cursor: pointer;\n",
       "    color: var(--link-color);\n",
       "    text-decoration: none;\n",
       "}\n",
       "\n",
       "table.dataframe tr:hover > td a {\n",
       "    color: var(--link-color-pale);\n",
       "}\n",
       "\n",
       "table.dataframe a:hover {\n",
       "    color: var(--link-hover);\n",
       "    text-decoration: underline;\n",
       "}\n",
       "\n",
       "table.dataframe img {\n",
       "    max-width: fit-content;\n",
       "}\n",
       "\n",
       "table.dataframe th.complex {\n",
       "    background-color: var(--background);\n",
       "    border: 1px solid var(--background);\n",
       "}\n",
       "\n",
       "table.dataframe .leftBorder {\n",
       "    border-left-color: var(--inner-border-color);\n",
       "}\n",
       "\n",
       "table.dataframe .rightBorder {\n",
       "    border-right-color: var(--inner-border-color);\n",
       "}\n",
       "\n",
       "table.dataframe .rightAlign {\n",
       "    text-align: right;\n",
       "}\n",
       "\n",
       "table.dataframe .expanderSvg {\n",
       "    width: 8px;\n",
       "    height: 8px;\n",
       "    margin-right: 3px;\n",
       "}\n",
       "\n",
       "table.dataframe .expander {\n",
       "    display: flex;\n",
       "    align-items: center;\n",
       "}\n",
       "\n",
       "/* formatting */\n",
       "\n",
       "table.dataframe .null {\n",
       "    color: var(--text-color-pale);\n",
       "}\n",
       "\n",
       "table.dataframe .structural {\n",
       "    color: var(--text-color-medium);\n",
       "    font-weight: bold;\n",
       "}\n",
       "\n",
       "table.dataframe .dataFrameCaption {\n",
       "    font-weight: bold;\n",
       "}\n",
       "\n",
       "table.dataframe .numbers {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "table.dataframe td:hover .formatted .structural, .null {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "table.dataframe tr:hover .formatted .structural, .null {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "\n",
       "\n",
       "\n",
       "            </style>\n",
       "        </head>\n",
       "        <body>\n",
       "            \n",
       "<table class=\"dataframe\" id=\"df_-2046820352\"></table>\n",
       "\n",
       "<p class=\"dataframe_description\">DataFrame: rowsCount = 10, columnsCount = 2</p>\n",
       "        </body>\n",
       "        <script>\n",
       "            \n",
       "/*<!--*/\n",
       "call_DataFrame(function() { DataFrame.addTable({ cols: [{ name: \"<span title=\\\"letter: String\\\">letter</span>\", children: [], rightAlign: false, values: [\"B\",\"A\",\"B\",\"C\",\"C\",\"B\",\"C\",\"B\",\"A\",\"A\"] }, \n",
       "{ name: \"<span title=\\\"num: Double\\\">num</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">0.624629</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">-0.858192</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">0.676221</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1.126394</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1.437662</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1.492079</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">-1.598537</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">-0.077561</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1.991766</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">0.270969</span></span>\"] }, \n",
       "], id: -2046820352, rootId: -2046820352, totalRows: 10 } ) });\n",
       "/*-->*/\n",
       "\n",
       "call_DataFrame(function() { DataFrame.renderTable(-2046820352) });\n",
       "\n",
       "\n",
       "        </script>\n",
       "        </html>"
      ]
     },
     "execution_count": 2,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "val rand = java.util.Random(100)\n",
    "val letters = sequenceOf(\"A\", \"B\", \"C\")\n",
    "\n",
    "val df = mapOf(\n",
    "    \"letter\" to List(10) { letters.shuffled().find { true } },\n",
    "    \"num\" to List(10) { rand.nextGaussian() }\n",
    ").toDataFrame()\n",
    "\n",
    "df"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "f7d81e8e-9e50-40c0-a45f-1152a7675ccf",
   "metadata": {},
   "source": [
    "If we call a [`groupBy`](https://kotlin.github.io/dataframe/groupby.html) with the `letter` column, we will *split* the dataset into three groups, one for each letter. In Kotlin DataFrame, we can print easily the content of the grouping operation, whereas in python we have to loop through the content of each group, because calling `groupby` will not compute anything except some intermediate data about the group key."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 3,
   "id": "c104ea74-076c-4aac-9b37-128a4a348bc5",
   "metadata": {
    "tags": []
   },
   "outputs": [
    {
     "data": {
      "application/kotlindataframe+json": "{\"nrow\":3,\"ncol\":2,\"columns\":[\"letter\",\"group\"],\"kotlin_dataframe\":[{\"letter\":\"B\",\"group\":[{\"letter\":\"B\",\"num\":0.6246292191371761},{\"letter\":\"B\",\"num\":0.6762208162903859},{\"letter\":\"B\",\"num\":1.4920787730184157},{\"letter\":\"B\",\"num\":-0.07756122563035872}]},{\"letter\":\"A\",\"group\":[{\"letter\":\"A\",\"num\":-0.8581918080882499},{\"letter\":\"A\",\"num\":1.9917663922907851},{\"letter\":\"A\",\"num\":0.27096900216864345}]},{\"letter\":\"C\",\"group\":[{\"letter\":\"C\",\"num\":1.126393826325953},{\"letter\":\"C\",\"num\":1.4376621993807677},{\"letter\":\"C\",\"num\":-1.5985374620272976}]}]}",
      "text/html": [
       "        <html>\n",
       "        <head>\n",
       "            <style type=\"text/css\">\n",
       "                :root {\n",
       "    --background: #fff;\n",
       "    --background-odd: #f5f5f5;\n",
       "    --background-hover: #d9edfd;\n",
       "    --header-text-color: #474747;\n",
       "    --text-color: #848484;\n",
       "    --text-color-dark: #000;\n",
       "    --text-color-medium: #737373;\n",
       "    --text-color-pale: #b3b3b3;\n",
       "    --inner-border-color: #aaa;\n",
       "    --bold-border-color: #000;\n",
       "    --link-color: #296eaa;\n",
       "    --link-color-pale: #296eaa;\n",
       "    --link-hover: #1a466c;\n",
       "}\n",
       "\n",
       ":root[theme=\"dark\"], :root [data-jp-theme-light=\"false\"], .dataframe_dark{\n",
       "    --background: #303030;\n",
       "    --background-odd: #3c3c3c;\n",
       "    --background-hover: #464646;\n",
       "    --header-text-color: #dddddd;\n",
       "    --text-color: #b3b3b3;\n",
       "    --text-color-dark: #dddddd;\n",
       "    --text-color-medium: #b2b2b2;\n",
       "    --text-color-pale: #737373;\n",
       "    --inner-border-color: #707070;\n",
       "    --bold-border-color: #777777;\n",
       "    --link-color: #008dc0;\n",
       "    --link-color-pale: #97e1fb;\n",
       "    --link-hover: #00688e;\n",
       "}\n",
       "\n",
       "p.dataframe_description {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "table.dataframe {\n",
       "    font-family: \"Helvetica Neue\", Helvetica, Arial, sans-serif;\n",
       "    font-size: 12px;\n",
       "    background-color: var(--background);\n",
       "    color: var(--text-color-dark);\n",
       "    border: none;\n",
       "    border-collapse: collapse;\n",
       "}\n",
       "\n",
       "table.dataframe th, td {\n",
       "    padding: 6px;\n",
       "    border: 1px solid transparent;\n",
       "    text-align: left;\n",
       "}\n",
       "\n",
       "table.dataframe th {\n",
       "    background-color: var(--background);\n",
       "    color: var(--header-text-color);\n",
       "}\n",
       "\n",
       "table.dataframe td {\n",
       "    vertical-align: top;\n",
       "}\n",
       "\n",
       "table.dataframe th.bottomBorder {\n",
       "    border-bottom-color: var(--bold-border-color);\n",
       "}\n",
       "\n",
       "table.dataframe tbody > tr:nth-child(odd) {\n",
       "    background: var(--background-odd);\n",
       "}\n",
       "\n",
       "table.dataframe tbody > tr:nth-child(even) {\n",
       "    background: var(--background);\n",
       "}\n",
       "\n",
       "table.dataframe tbody > tr:hover {\n",
       "    background: var(--background-hover);\n",
       "}\n",
       "\n",
       "table.dataframe a {\n",
       "    cursor: pointer;\n",
       "    color: var(--link-color);\n",
       "    text-decoration: none;\n",
       "}\n",
       "\n",
       "table.dataframe tr:hover > td a {\n",
       "    color: var(--link-color-pale);\n",
       "}\n",
       "\n",
       "table.dataframe a:hover {\n",
       "    color: var(--link-hover);\n",
       "    text-decoration: underline;\n",
       "}\n",
       "\n",
       "table.dataframe img {\n",
       "    max-width: fit-content;\n",
       "}\n",
       "\n",
       "table.dataframe th.complex {\n",
       "    background-color: var(--background);\n",
       "    border: 1px solid var(--background);\n",
       "}\n",
       "\n",
       "table.dataframe .leftBorder {\n",
       "    border-left-color: var(--inner-border-color);\n",
       "}\n",
       "\n",
       "table.dataframe .rightBorder {\n",
       "    border-right-color: var(--inner-border-color);\n",
       "}\n",
       "\n",
       "table.dataframe .rightAlign {\n",
       "    text-align: right;\n",
       "}\n",
       "\n",
       "table.dataframe .expanderSvg {\n",
       "    width: 8px;\n",
       "    height: 8px;\n",
       "    margin-right: 3px;\n",
       "}\n",
       "\n",
       "table.dataframe .expander {\n",
       "    display: flex;\n",
       "    align-items: center;\n",
       "}\n",
       "\n",
       "/* formatting */\n",
       "\n",
       "table.dataframe .null {\n",
       "    color: var(--text-color-pale);\n",
       "}\n",
       "\n",
       "table.dataframe .structural {\n",
       "    color: var(--text-color-medium);\n",
       "    font-weight: bold;\n",
       "}\n",
       "\n",
       "table.dataframe .dataFrameCaption {\n",
       "    font-weight: bold;\n",
       "}\n",
       "\n",
       "table.dataframe .numbers {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "table.dataframe td:hover .formatted .structural, .null {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "table.dataframe tr:hover .formatted .structural, .null {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "\n",
       "\n",
       "\n",
       "            </style>\n",
       "        </head>\n",
       "        <body>\n",
       "            \n",
       "<table class=\"dataframe\" id=\"df_-2046820351\"></table>\n",
       "\n",
       "<p class=\"dataframe_description\">GroupBy</p>\n",
       "        </body>\n",
       "        <script>\n",
       "            \n",
       "/*<!--*/\n",
       "call_DataFrame(function() { DataFrame.addTable({ cols: [{ name: \"<span title=\\\"letter: String\\\">letter</span>\", children: [], rightAlign: false, values: [\"B\",\"A\",\"C\"] }, \n",
       "{ name: \"<span title=\\\"group: DataFrame\\\">group</span>\", children: [], rightAlign: false, values: [{ frameId: -2046820350, value: \"<b>DataFrame 4 x 2</b>\" },{ frameId: -2046820349, value: \"<b>DataFrame 3 x 2</b>\" },{ frameId: -2046820348, value: \"<b>DataFrame 3 x 2</b>\" }] }, \n",
       "], id: -2046820351, rootId: -2046820351, totalRows: 3 } ) });\n",
       "/*-->*/\n",
       "\n",
       "/*<!--*/\n",
       "call_DataFrame(function() { DataFrame.addTable({ cols: [{ name: \"<span title=\\\"letter: String\\\">letter</span>\", children: [], rightAlign: false, values: [\"B\",\"B\",\"B\",\"B\"] }, \n",
       "{ name: \"<span title=\\\"num: Double\\\">num</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">0.624629</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">0.676221</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1.492079</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">-0.077561</span></span>\"] }, \n",
       "], id: -2046820350, rootId: -2046820351, totalRows: 4 } ) });\n",
       "/*-->*/\n",
       "\n",
       "/*<!--*/\n",
       "call_DataFrame(function() { DataFrame.addTable({ cols: [{ name: \"<span title=\\\"letter: String\\\">letter</span>\", children: [], rightAlign: false, values: [\"A\",\"A\",\"A\"] }, \n",
       "{ name: \"<span title=\\\"num: Double\\\">num</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">-0.858192</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1.991766</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">0.270969</span></span>\"] }, \n",
       "], id: -2046820349, rootId: -2046820351, totalRows: 3 } ) });\n",
       "/*-->*/\n",
       "\n",
       "/*<!--*/\n",
       "call_DataFrame(function() { DataFrame.addTable({ cols: [{ name: \"<span title=\\\"letter: String\\\">letter</span>\", children: [], rightAlign: false, values: [\"C\",\"C\",\"C\"] }, \n",
       "{ name: \"<span title=\\\"num: Double\\\">num</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1.126394</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1.437662</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">-1.598537</span></span>\"] }, \n",
       "], id: -2046820348, rootId: -2046820351, totalRows: 3 } ) });\n",
       "/*-->*/\n",
       "\n",
       "call_DataFrame(function() { DataFrame.renderTable(-2046820351) });\n",
       "\n",
       "\n",
       "        </script>\n",
       "        </html>"
      ]
     },
     "execution_count": 3,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "df.groupBy { letter }"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "16036a4a-8ba5-422f-80ad-7ab4afdc1e8e",
   "metadata": {},
   "source": [
    "And we can see all the groups that are being formed.\n",
    "The `groupBy` method accept any column expression provided, so the following will be still legit:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 4,
   "id": "626a2906-367c-47a6-9d48-03e1cf3357a9",
   "metadata": {
    "tags": []
   },
   "outputs": [
    {
     "data": {
      "application/kotlindataframe+json": "{\"nrow\":2,\"ncol\":2,\"columns\":[\"letter\",\"group\"],\"kotlin_dataframe\":[{\"letter\":false,\"group\":[{\"letter\":\"B\",\"num\":0.6246292191371761},{\"letter\":\"B\",\"num\":0.6762208162903859},{\"letter\":\"C\",\"num\":1.126393826325953},{\"letter\":\"C\",\"num\":1.4376621993807677},{\"letter\":\"B\",\"num\":1.4920787730184157},{\"letter\":\"C\",\"num\":-1.5985374620272976},{\"letter\":\"B\",\"num\":-0.07756122563035872}]},{\"letter\":true,\"group\":[{\"letter\":\"A\",\"num\":-0.8581918080882499},{\"letter\":\"A\",\"num\":1.9917663922907851},{\"letter\":\"A\",\"num\":0.27096900216864345}]}]}",
      "text/html": [
       "        <html>\n",
       "        <head>\n",
       "            <style type=\"text/css\">\n",
       "                :root {\n",
       "    --background: #fff;\n",
       "    --background-odd: #f5f5f5;\n",
       "    --background-hover: #d9edfd;\n",
       "    --header-text-color: #474747;\n",
       "    --text-color: #848484;\n",
       "    --text-color-dark: #000;\n",
       "    --text-color-medium: #737373;\n",
       "    --text-color-pale: #b3b3b3;\n",
       "    --inner-border-color: #aaa;\n",
       "    --bold-border-color: #000;\n",
       "    --link-color: #296eaa;\n",
       "    --link-color-pale: #296eaa;\n",
       "    --link-hover: #1a466c;\n",
       "}\n",
       "\n",
       ":root[theme=\"dark\"], :root [data-jp-theme-light=\"false\"], .dataframe_dark{\n",
       "    --background: #303030;\n",
       "    --background-odd: #3c3c3c;\n",
       "    --background-hover: #464646;\n",
       "    --header-text-color: #dddddd;\n",
       "    --text-color: #b3b3b3;\n",
       "    --text-color-dark: #dddddd;\n",
       "    --text-color-medium: #b2b2b2;\n",
       "    --text-color-pale: #737373;\n",
       "    --inner-border-color: #707070;\n",
       "    --bold-border-color: #777777;\n",
       "    --link-color: #008dc0;\n",
       "    --link-color-pale: #97e1fb;\n",
       "    --link-hover: #00688e;\n",
       "}\n",
       "\n",
       "p.dataframe_description {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "table.dataframe {\n",
       "    font-family: \"Helvetica Neue\", Helvetica, Arial, sans-serif;\n",
       "    font-size: 12px;\n",
       "    background-color: var(--background);\n",
       "    color: var(--text-color-dark);\n",
       "    border: none;\n",
       "    border-collapse: collapse;\n",
       "}\n",
       "\n",
       "table.dataframe th, td {\n",
       "    padding: 6px;\n",
       "    border: 1px solid transparent;\n",
       "    text-align: left;\n",
       "}\n",
       "\n",
       "table.dataframe th {\n",
       "    background-color: var(--background);\n",
       "    color: var(--header-text-color);\n",
       "}\n",
       "\n",
       "table.dataframe td {\n",
       "    vertical-align: top;\n",
       "}\n",
       "\n",
       "table.dataframe th.bottomBorder {\n",
       "    border-bottom-color: var(--bold-border-color);\n",
       "}\n",
       "\n",
       "table.dataframe tbody > tr:nth-child(odd) {\n",
       "    background: var(--background-odd);\n",
       "}\n",
       "\n",
       "table.dataframe tbody > tr:nth-child(even) {\n",
       "    background: var(--background);\n",
       "}\n",
       "\n",
       "table.dataframe tbody > tr:hover {\n",
       "    background: var(--background-hover);\n",
       "}\n",
       "\n",
       "table.dataframe a {\n",
       "    cursor: pointer;\n",
       "    color: var(--link-color);\n",
       "    text-decoration: none;\n",
       "}\n",
       "\n",
       "table.dataframe tr:hover > td a {\n",
       "    color: var(--link-color-pale);\n",
       "}\n",
       "\n",
       "table.dataframe a:hover {\n",
       "    color: var(--link-hover);\n",
       "    text-decoration: underline;\n",
       "}\n",
       "\n",
       "table.dataframe img {\n",
       "    max-width: fit-content;\n",
       "}\n",
       "\n",
       "table.dataframe th.complex {\n",
       "    background-color: var(--background);\n",
       "    border: 1px solid var(--background);\n",
       "}\n",
       "\n",
       "table.dataframe .leftBorder {\n",
       "    border-left-color: var(--inner-border-color);\n",
       "}\n",
       "\n",
       "table.dataframe .rightBorder {\n",
       "    border-right-color: var(--inner-border-color);\n",
       "}\n",
       "\n",
       "table.dataframe .rightAlign {\n",
       "    text-align: right;\n",
       "}\n",
       "\n",
       "table.dataframe .expanderSvg {\n",
       "    width: 8px;\n",
       "    height: 8px;\n",
       "    margin-right: 3px;\n",
       "}\n",
       "\n",
       "table.dataframe .expander {\n",
       "    display: flex;\n",
       "    align-items: center;\n",
       "}\n",
       "\n",
       "/* formatting */\n",
       "\n",
       "table.dataframe .null {\n",
       "    color: var(--text-color-pale);\n",
       "}\n",
       "\n",
       "table.dataframe .structural {\n",
       "    color: var(--text-color-medium);\n",
       "    font-weight: bold;\n",
       "}\n",
       "\n",
       "table.dataframe .dataFrameCaption {\n",
       "    font-weight: bold;\n",
       "}\n",
       "\n",
       "table.dataframe .numbers {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "table.dataframe td:hover .formatted .structural, .null {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "table.dataframe tr:hover .formatted .structural, .null {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "\n",
       "\n",
       "\n",
       "            </style>\n",
       "        </head>\n",
       "        <body>\n",
       "            \n",
       "<table class=\"dataframe\" id=\"df_-2046820347\"></table>\n",
       "\n",
       "<p class=\"dataframe_description\">GroupBy</p>\n",
       "        </body>\n",
       "        <script>\n",
       "            \n",
       "/*<!--*/\n",
       "call_DataFrame(function() { DataFrame.addTable({ cols: [{ name: \"<span title=\\\"letter: Boolean\\\">letter</span>\", children: [], rightAlign: false, values: [\"false\",\"true\"] }, \n",
       "{ name: \"<span title=\\\"group: DataFrame\\\">group</span>\", children: [], rightAlign: false, values: [{ frameId: -2046820346, value: \"<b>DataFrame 7 x 2</b>\" },{ frameId: -2046820345, value: \"<b>DataFrame 3 x 2</b>\" }] }, \n",
       "], id: -2046820347, rootId: -2046820347, totalRows: 2 } ) });\n",
       "/*-->*/\n",
       "\n",
       "/*<!--*/\n",
       "call_DataFrame(function() { DataFrame.addTable({ cols: [{ name: \"<span title=\\\"letter: String\\\">letter</span>\", children: [], rightAlign: false, values: [\"B\",\"B\",\"C\",\"C\",\"B\"] }, \n",
       "{ name: \"<span title=\\\"num: Double\\\">num</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">0.624629</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">0.676221</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1.126394</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1.437662</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1.492079</span></span>\"] }, \n",
       "], id: -2046820346, rootId: -2046820347, totalRows: 7 } ) });\n",
       "/*-->*/\n",
       "\n",
       "/*<!--*/\n",
       "call_DataFrame(function() { DataFrame.addTable({ cols: [{ name: \"<span title=\\\"letter: String\\\">letter</span>\", children: [], rightAlign: false, values: [\"A\",\"A\",\"A\"] }, \n",
       "{ name: \"<span title=\\\"num: Double\\\">num</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">-0.858192</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1.991766</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">0.270969</span></span>\"] }, \n",
       "], id: -2046820345, rootId: -2046820347, totalRows: 3 } ) });\n",
       "/*-->*/\n",
       "\n",
       "call_DataFrame(function() { DataFrame.renderTable(-2046820347) });\n",
       "\n",
       "\n",
       "        </script>\n",
       "        </html>"
      ]
     },
     "execution_count": 4,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "df.groupBy { letter.map { it == \"A\"} }"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "13d392c6-50a9-4ade-b07d-45dc51d8439d",
   "metadata": {},
   "source": [
    "This is very useful when manipulating `date` objects and we want to group by year/month/day."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 5,
   "id": "f78e520c-10ac-4490-a87a-90c9fb11d6ef",
   "metadata": {
    "tags": []
   },
   "outputs": [],
   "source": [
    "val dates = listOf(\n",
    "    LocalDate(1998, 1, 2),\n",
    "    LocalDate(1999, 11, 21),\n",
    "    LocalDate(1999, 3, 12),\n",
    "    LocalDate(1998, 6, 22),\n",
    "    LocalDate(1998, 12, 25),\n",
    "    LocalDate(1999, 12, 24),\n",
    "    LocalDate(1999, 2, 9),\n",
    "    LocalDate(1999, 6, 24),\n",
    "    LocalDate(1998, 1, 20),\n",
    "    LocalDate(1999, 12, 20),\n",
    ").toColumn(\"date\")\n",
    "\n",
    "val dfd = df.add(dates)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 6,
   "id": "09e2f337-96a0-42f6-931a-ce93ba6a38c4",
   "metadata": {
    "tags": []
   },
   "outputs": [
    {
     "data": {
      "application/kotlindataframe+json": "{\"nrow\":2,\"ncol\":2,\"columns\":[\"date\",\"group\"],\"kotlin_dataframe\":[{\"date\":1998,\"group\":[{\"letter\":\"B\",\"num\":0.6246292191371761,\"date\":\"1998-01-02\"},{\"letter\":\"C\",\"num\":1.126393826325953,\"date\":\"1998-06-22\"},{\"letter\":\"C\",\"num\":1.4376621993807677,\"date\":\"1998-12-25\"},{\"letter\":\"A\",\"num\":1.9917663922907851,\"date\":\"1998-01-20\"}]},{\"date\":1999,\"group\":[{\"letter\":\"A\",\"num\":-0.8581918080882499,\"date\":\"1999-11-21\"},{\"letter\":\"B\",\"num\":0.6762208162903859,\"date\":\"1999-03-12\"},{\"letter\":\"B\",\"num\":1.4920787730184157,\"date\":\"1999-12-24\"},{\"letter\":\"C\",\"num\":-1.5985374620272976,\"date\":\"1999-02-09\"},{\"letter\":\"B\",\"num\":-0.07756122563035872,\"date\":\"1999-06-24\"},{\"letter\":\"A\",\"num\":0.27096900216864345,\"date\":\"1999-12-20\"}]}]}",
      "text/html": [
       "        <html>\n",
       "        <head>\n",
       "            <style type=\"text/css\">\n",
       "                :root {\n",
       "    --background: #fff;\n",
       "    --background-odd: #f5f5f5;\n",
       "    --background-hover: #d9edfd;\n",
       "    --header-text-color: #474747;\n",
       "    --text-color: #848484;\n",
       "    --text-color-dark: #000;\n",
       "    --text-color-medium: #737373;\n",
       "    --text-color-pale: #b3b3b3;\n",
       "    --inner-border-color: #aaa;\n",
       "    --bold-border-color: #000;\n",
       "    --link-color: #296eaa;\n",
       "    --link-color-pale: #296eaa;\n",
       "    --link-hover: #1a466c;\n",
       "}\n",
       "\n",
       ":root[theme=\"dark\"], :root [data-jp-theme-light=\"false\"], .dataframe_dark{\n",
       "    --background: #303030;\n",
       "    --background-odd: #3c3c3c;\n",
       "    --background-hover: #464646;\n",
       "    --header-text-color: #dddddd;\n",
       "    --text-color: #b3b3b3;\n",
       "    --text-color-dark: #dddddd;\n",
       "    --text-color-medium: #b2b2b2;\n",
       "    --text-color-pale: #737373;\n",
       "    --inner-border-color: #707070;\n",
       "    --bold-border-color: #777777;\n",
       "    --link-color: #008dc0;\n",
       "    --link-color-pale: #97e1fb;\n",
       "    --link-hover: #00688e;\n",
       "}\n",
       "\n",
       "p.dataframe_description {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "table.dataframe {\n",
       "    font-family: \"Helvetica Neue\", Helvetica, Arial, sans-serif;\n",
       "    font-size: 12px;\n",
       "    background-color: var(--background);\n",
       "    color: var(--text-color-dark);\n",
       "    border: none;\n",
       "    border-collapse: collapse;\n",
       "}\n",
       "\n",
       "table.dataframe th, td {\n",
       "    padding: 6px;\n",
       "    border: 1px solid transparent;\n",
       "    text-align: left;\n",
       "}\n",
       "\n",
       "table.dataframe th {\n",
       "    background-color: var(--background);\n",
       "    color: var(--header-text-color);\n",
       "}\n",
       "\n",
       "table.dataframe td {\n",
       "    vertical-align: top;\n",
       "}\n",
       "\n",
       "table.dataframe th.bottomBorder {\n",
       "    border-bottom-color: var(--bold-border-color);\n",
       "}\n",
       "\n",
       "table.dataframe tbody > tr:nth-child(odd) {\n",
       "    background: var(--background-odd);\n",
       "}\n",
       "\n",
       "table.dataframe tbody > tr:nth-child(even) {\n",
       "    background: var(--background);\n",
       "}\n",
       "\n",
       "table.dataframe tbody > tr:hover {\n",
       "    background: var(--background-hover);\n",
       "}\n",
       "\n",
       "table.dataframe a {\n",
       "    cursor: pointer;\n",
       "    color: var(--link-color);\n",
       "    text-decoration: none;\n",
       "}\n",
       "\n",
       "table.dataframe tr:hover > td a {\n",
       "    color: var(--link-color-pale);\n",
       "}\n",
       "\n",
       "table.dataframe a:hover {\n",
       "    color: var(--link-hover);\n",
       "    text-decoration: underline;\n",
       "}\n",
       "\n",
       "table.dataframe img {\n",
       "    max-width: fit-content;\n",
       "}\n",
       "\n",
       "table.dataframe th.complex {\n",
       "    background-color: var(--background);\n",
       "    border: 1px solid var(--background);\n",
       "}\n",
       "\n",
       "table.dataframe .leftBorder {\n",
       "    border-left-color: var(--inner-border-color);\n",
       "}\n",
       "\n",
       "table.dataframe .rightBorder {\n",
       "    border-right-color: var(--inner-border-color);\n",
       "}\n",
       "\n",
       "table.dataframe .rightAlign {\n",
       "    text-align: right;\n",
       "}\n",
       "\n",
       "table.dataframe .expanderSvg {\n",
       "    width: 8px;\n",
       "    height: 8px;\n",
       "    margin-right: 3px;\n",
       "}\n",
       "\n",
       "table.dataframe .expander {\n",
       "    display: flex;\n",
       "    align-items: center;\n",
       "}\n",
       "\n",
       "/* formatting */\n",
       "\n",
       "table.dataframe .null {\n",
       "    color: var(--text-color-pale);\n",
       "}\n",
       "\n",
       "table.dataframe .structural {\n",
       "    color: var(--text-color-medium);\n",
       "    font-weight: bold;\n",
       "}\n",
       "\n",
       "table.dataframe .dataFrameCaption {\n",
       "    font-weight: bold;\n",
       "}\n",
       "\n",
       "table.dataframe .numbers {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "table.dataframe td:hover .formatted .structural, .null {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "table.dataframe tr:hover .formatted .structural, .null {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "\n",
       "\n",
       "\n",
       "            </style>\n",
       "        </head>\n",
       "        <body>\n",
       "            \n",
       "<table class=\"dataframe\" id=\"df_-2046820344\"></table>\n",
       "\n",
       "<p class=\"dataframe_description\">GroupBy</p>\n",
       "        </body>\n",
       "        <script>\n",
       "            \n",
       "/*<!--*/\n",
       "call_DataFrame(function() { DataFrame.addTable({ cols: [{ name: \"<span title=\\\"date: Int\\\">date</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1998</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1999</span></span>\"] }, \n",
       "{ name: \"<span title=\\\"group: DataFrame\\\">group</span>\", children: [], rightAlign: false, values: [{ frameId: -2046820343, value: \"<b>DataFrame 4 x 3</b>\" },{ frameId: -2046820342, value: \"<b>DataFrame 6 x 3</b>\" }] }, \n",
       "], id: -2046820344, rootId: -2046820344, totalRows: 2 } ) });\n",
       "/*-->*/\n",
       "\n",
       "/*<!--*/\n",
       "call_DataFrame(function() { DataFrame.addTable({ cols: [{ name: \"<span title=\\\"letter: String\\\">letter</span>\", children: [], rightAlign: false, values: [\"B\",\"C\",\"C\",\"A\"] }, \n",
       "{ name: \"<span title=\\\"num: Double\\\">num</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">0.624629</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1.126394</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1.437662</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1.991766</span></span>\"] }, \n",
       "{ name: \"<span title=\\\"date: kotlinx.datetime.LocalDate\\\">date</span>\", children: [], rightAlign: false, values: [\"1998-01-02\",\"1998-06-22\",\"1998-12-25\",\"1998-01-20\"] }, \n",
       "], id: -2046820343, rootId: -2046820344, totalRows: 4 } ) });\n",
       "/*-->*/\n",
       "\n",
       "/*<!--*/\n",
       "call_DataFrame(function() { DataFrame.addTable({ cols: [{ name: \"<span title=\\\"letter: String\\\">letter</span>\", children: [], rightAlign: false, values: [\"A\",\"B\",\"B\",\"C\",\"B\"] }, \n",
       "{ name: \"<span title=\\\"num: Double\\\">num</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">-0.858192</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">0.676221</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1.492079</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">-1.598537</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">-0.077561</span></span>\"] }, \n",
       "{ name: \"<span title=\\\"date: kotlinx.datetime.LocalDate\\\">date</span>\", children: [], rightAlign: false, values: [\"1999-11-21\",\"1999-03-12\",\"1999-12-24\",\"1999-02-09\",\"1999-06-24\"] }, \n",
       "], id: -2046820342, rootId: -2046820344, totalRows: 6 } ) });\n",
       "/*-->*/\n",
       "\n",
       "call_DataFrame(function() { DataFrame.renderTable(-2046820344) });\n",
       "\n",
       "\n",
       "        </script>\n",
       "        </html>"
      ]
     },
     "execution_count": 6,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "dfd.groupBy { date.map { it.year } }"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "d251eaf5-8ae4-4346-8a18-206dc4d66a42",
   "metadata": {},
   "source": [
    "`groupBy` accepts any number of column names to group by with."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 7,
   "id": "80de6e23-3979-48d1-96bd-8cdb4eebae4e",
   "metadata": {
    "tags": []
   },
   "outputs": [
    {
     "data": {
      "application/kotlindataframe+json": "{\"nrow\":6,\"ncol\":3,\"columns\":[\"letter\",\"date\",\"group\"],\"kotlin_dataframe\":[{\"letter\":\"B\",\"date\":1998,\"group\":[{\"letter\":\"B\",\"num\":0.6246292191371761,\"date\":\"1998-01-02\"}]},{\"letter\":\"A\",\"date\":1999,\"group\":[{\"letter\":\"A\",\"num\":-0.8581918080882499,\"date\":\"1999-11-21\"},{\"letter\":\"A\",\"num\":0.27096900216864345,\"date\":\"1999-12-20\"}]},{\"letter\":\"B\",\"date\":1999,\"group\":[{\"letter\":\"B\",\"num\":0.6762208162903859,\"date\":\"1999-03-12\"},{\"letter\":\"B\",\"num\":1.4920787730184157,\"date\":\"1999-12-24\"},{\"letter\":\"B\",\"num\":-0.07756122563035872,\"date\":\"1999-06-24\"}]},{\"letter\":\"C\",\"date\":1998,\"group\":[{\"letter\":\"C\",\"num\":1.126393826325953,\"date\":\"1998-06-22\"},{\"letter\":\"C\",\"num\":1.4376621993807677,\"date\":\"1998-12-25\"}]},{\"letter\":\"C\",\"date\":1999,\"group\":[{\"letter\":\"C\",\"num\":-1.5985374620272976,\"date\":\"1999-02-09\"}]},{\"letter\":\"A\",\"date\":1998,\"group\":[{\"letter\":\"A\",\"num\":1.9917663922907851,\"date\":\"1998-01-20\"}]}]}",
      "text/html": [
       "        <html>\n",
       "        <head>\n",
       "            <style type=\"text/css\">\n",
       "                :root {\n",
       "    --background: #fff;\n",
       "    --background-odd: #f5f5f5;\n",
       "    --background-hover: #d9edfd;\n",
       "    --header-text-color: #474747;\n",
       "    --text-color: #848484;\n",
       "    --text-color-dark: #000;\n",
       "    --text-color-medium: #737373;\n",
       "    --text-color-pale: #b3b3b3;\n",
       "    --inner-border-color: #aaa;\n",
       "    --bold-border-color: #000;\n",
       "    --link-color: #296eaa;\n",
       "    --link-color-pale: #296eaa;\n",
       "    --link-hover: #1a466c;\n",
       "}\n",
       "\n",
       ":root[theme=\"dark\"], :root [data-jp-theme-light=\"false\"], .dataframe_dark{\n",
       "    --background: #303030;\n",
       "    --background-odd: #3c3c3c;\n",
       "    --background-hover: #464646;\n",
       "    --header-text-color: #dddddd;\n",
       "    --text-color: #b3b3b3;\n",
       "    --text-color-dark: #dddddd;\n",
       "    --text-color-medium: #b2b2b2;\n",
       "    --text-color-pale: #737373;\n",
       "    --inner-border-color: #707070;\n",
       "    --bold-border-color: #777777;\n",
       "    --link-color: #008dc0;\n",
       "    --link-color-pale: #97e1fb;\n",
       "    --link-hover: #00688e;\n",
       "}\n",
       "\n",
       "p.dataframe_description {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "table.dataframe {\n",
       "    font-family: \"Helvetica Neue\", Helvetica, Arial, sans-serif;\n",
       "    font-size: 12px;\n",
       "    background-color: var(--background);\n",
       "    color: var(--text-color-dark);\n",
       "    border: none;\n",
       "    border-collapse: collapse;\n",
       "}\n",
       "\n",
       "table.dataframe th, td {\n",
       "    padding: 6px;\n",
       "    border: 1px solid transparent;\n",
       "    text-align: left;\n",
       "}\n",
       "\n",
       "table.dataframe th {\n",
       "    background-color: var(--background);\n",
       "    color: var(--header-text-color);\n",
       "}\n",
       "\n",
       "table.dataframe td {\n",
       "    vertical-align: top;\n",
       "}\n",
       "\n",
       "table.dataframe th.bottomBorder {\n",
       "    border-bottom-color: var(--bold-border-color);\n",
       "}\n",
       "\n",
       "table.dataframe tbody > tr:nth-child(odd) {\n",
       "    background: var(--background-odd);\n",
       "}\n",
       "\n",
       "table.dataframe tbody > tr:nth-child(even) {\n",
       "    background: var(--background);\n",
       "}\n",
       "\n",
       "table.dataframe tbody > tr:hover {\n",
       "    background: var(--background-hover);\n",
       "}\n",
       "\n",
       "table.dataframe a {\n",
       "    cursor: pointer;\n",
       "    color: var(--link-color);\n",
       "    text-decoration: none;\n",
       "}\n",
       "\n",
       "table.dataframe tr:hover > td a {\n",
       "    color: var(--link-color-pale);\n",
       "}\n",
       "\n",
       "table.dataframe a:hover {\n",
       "    color: var(--link-hover);\n",
       "    text-decoration: underline;\n",
       "}\n",
       "\n",
       "table.dataframe img {\n",
       "    max-width: fit-content;\n",
       "}\n",
       "\n",
       "table.dataframe th.complex {\n",
       "    background-color: var(--background);\n",
       "    border: 1px solid var(--background);\n",
       "}\n",
       "\n",
       "table.dataframe .leftBorder {\n",
       "    border-left-color: var(--inner-border-color);\n",
       "}\n",
       "\n",
       "table.dataframe .rightBorder {\n",
       "    border-right-color: var(--inner-border-color);\n",
       "}\n",
       "\n",
       "table.dataframe .rightAlign {\n",
       "    text-align: right;\n",
       "}\n",
       "\n",
       "table.dataframe .expanderSvg {\n",
       "    width: 8px;\n",
       "    height: 8px;\n",
       "    margin-right: 3px;\n",
       "}\n",
       "\n",
       "table.dataframe .expander {\n",
       "    display: flex;\n",
       "    align-items: center;\n",
       "}\n",
       "\n",
       "/* formatting */\n",
       "\n",
       "table.dataframe .null {\n",
       "    color: var(--text-color-pale);\n",
       "}\n",
       "\n",
       "table.dataframe .structural {\n",
       "    color: var(--text-color-medium);\n",
       "    font-weight: bold;\n",
       "}\n",
       "\n",
       "table.dataframe .dataFrameCaption {\n",
       "    font-weight: bold;\n",
       "}\n",
       "\n",
       "table.dataframe .numbers {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "table.dataframe td:hover .formatted .structural, .null {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "table.dataframe tr:hover .formatted .structural, .null {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "\n",
       "\n",
       "\n",
       "            </style>\n",
       "        </head>\n",
       "        <body>\n",
       "            \n",
       "<table class=\"dataframe\" id=\"df_-2046820341\"></table>\n",
       "\n",
       "<p class=\"dataframe_description\">GroupBy</p>\n",
       "        </body>\n",
       "        <script>\n",
       "            \n",
       "/*<!--*/\n",
       "call_DataFrame(function() { DataFrame.addTable({ cols: [{ name: \"<span title=\\\"letter: String\\\">letter</span>\", children: [], rightAlign: false, values: [\"B\",\"A\",\"B\",\"C\",\"C\",\"A\"] }, \n",
       "{ name: \"<span title=\\\"date: Int\\\">date</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1998</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1999</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1999</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1998</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1999</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1998</span></span>\"] }, \n",
       "{ name: \"<span title=\\\"group: DataFrame\\\">group</span>\", children: [], rightAlign: false, values: [{ frameId: -2046820340, value: \"<b>DataFrame 1 x 3</b>\" },{ frameId: -2046820339, value: \"<b>DataFrame 2 x 3</b>\" },{ frameId: -2046820338, value: \"<b>DataFrame 3 x 3</b>\" },{ frameId: -2046820337, value: \"<b>DataFrame 2 x 3</b>\" },{ frameId: -2046820336, value: \"<b>DataFrame 1 x 3</b>\" },{ frameId: -2046820335, value: \"<b>DataFrame 1 x 3</b>\" }] }, \n",
       "], id: -2046820341, rootId: -2046820341, totalRows: 6 } ) });\n",
       "/*-->*/\n",
       "\n",
       "/*<!--*/\n",
       "call_DataFrame(function() { DataFrame.addTable({ cols: [{ name: \"<span title=\\\"letter: String\\\">letter</span>\", children: [], rightAlign: false, values: [\"B\"] }, \n",
       "{ name: \"<span title=\\\"num: Double\\\">num</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">0.624629</span></span>\"] }, \n",
       "{ name: \"<span title=\\\"date: kotlinx.datetime.LocalDate\\\">date</span>\", children: [], rightAlign: false, values: [\"1998-01-02\"] }, \n",
       "], id: -2046820340, rootId: -2046820341, totalRows: 1 } ) });\n",
       "/*-->*/\n",
       "\n",
       "/*<!--*/\n",
       "call_DataFrame(function() { DataFrame.addTable({ cols: [{ name: \"<span title=\\\"letter: String\\\">letter</span>\", children: [], rightAlign: false, values: [\"A\",\"A\"] }, \n",
       "{ name: \"<span title=\\\"num: Double\\\">num</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">-0.858192</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">0.270969</span></span>\"] }, \n",
       "{ name: \"<span title=\\\"date: kotlinx.datetime.LocalDate\\\">date</span>\", children: [], rightAlign: false, values: [\"1999-11-21\",\"1999-12-20\"] }, \n",
       "], id: -2046820339, rootId: -2046820341, totalRows: 2 } ) });\n",
       "/*-->*/\n",
       "\n",
       "/*<!--*/\n",
       "call_DataFrame(function() { DataFrame.addTable({ cols: [{ name: \"<span title=\\\"letter: String\\\">letter</span>\", children: [], rightAlign: false, values: [\"B\",\"B\",\"B\"] }, \n",
       "{ name: \"<span title=\\\"num: Double\\\">num</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">0.676221</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1.492079</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">-0.077561</span></span>\"] }, \n",
       "{ name: \"<span title=\\\"date: kotlinx.datetime.LocalDate\\\">date</span>\", children: [], rightAlign: false, values: [\"1999-03-12\",\"1999-12-24\",\"1999-06-24\"] }, \n",
       "], id: -2046820338, rootId: -2046820341, totalRows: 3 } ) });\n",
       "/*-->*/\n",
       "\n",
       "/*<!--*/\n",
       "call_DataFrame(function() { DataFrame.addTable({ cols: [{ name: \"<span title=\\\"letter: String\\\">letter</span>\", children: [], rightAlign: false, values: [\"C\",\"C\"] }, \n",
       "{ name: \"<span title=\\\"num: Double\\\">num</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1.126394</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1.437662</span></span>\"] }, \n",
       "{ name: \"<span title=\\\"date: kotlinx.datetime.LocalDate\\\">date</span>\", children: [], rightAlign: false, values: [\"1998-06-22\",\"1998-12-25\"] }, \n",
       "], id: -2046820337, rootId: -2046820341, totalRows: 2 } ) });\n",
       "/*-->*/\n",
       "\n",
       "/*<!--*/\n",
       "call_DataFrame(function() { DataFrame.addTable({ cols: [{ name: \"<span title=\\\"letter: String\\\">letter</span>\", children: [], rightAlign: false, values: [\"C\"] }, \n",
       "{ name: \"<span title=\\\"num: Double\\\">num</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">-1.598537</span></span>\"] }, \n",
       "{ name: \"<span title=\\\"date: kotlinx.datetime.LocalDate\\\">date</span>\", children: [], rightAlign: false, values: [\"1999-02-09\"] }, \n",
       "], id: -2046820336, rootId: -2046820341, totalRows: 1 } ) });\n",
       "/*-->*/\n",
       "\n",
       "/*<!--*/\n",
       "call_DataFrame(function() { DataFrame.addTable({ cols: [{ name: \"<span title=\\\"letter: String\\\">letter</span>\", children: [], rightAlign: false, values: [\"A\"] }, \n",
       "{ name: \"<span title=\\\"num: Double\\\">num</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1.991766</span></span>\"] }, \n",
       "{ name: \"<span title=\\\"date: kotlinx.datetime.LocalDate\\\">date</span>\", children: [], rightAlign: false, values: [\"1998-01-20\"] }, \n",
       "], id: -2046820335, rootId: -2046820341, totalRows: 1 } ) });\n",
       "/*-->*/\n",
       "\n",
       "call_DataFrame(function() { DataFrame.renderTable(-2046820341) });\n",
       "\n",
       "\n",
       "        </script>\n",
       "        </html>"
      ]
     },
     "execution_count": 7,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "dfd.groupBy { letter and date.map { it.year } }"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "e97a8135-499a-4395-af34-2968b6130392",
   "metadata": {},
   "source": [
    "We can also override the name of the group key column in the resulting dataframe using the `named` keyword."
   ]
  },
  {
   "cell_type": "markdown",
   "id": "bf05af37-10ec-4129-a6ec-ce6b12fab44c",
   "metadata": {},
   "source": [
    "Before computing aggregation or reduction operations, we can apply some **transformation** to the groups, like sorting, or computing other groups modifications operations. Among those methods, there's the `concat` method, that unions all data groups of `groupBy` into original `DataFrame` preserving new order of rows produced by grouping.\n",
    "\n",
    "A very common pattern, is to use the concatenation of `groupBy` and `pivot` operations:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 8,
   "id": "8a193858-7abc-4ea2-8a19-5717ce6cf3bc",
   "metadata": {
    "tags": []
   },
   "outputs": [
    {
     "data": {
      "application/kotlindataframe+json": "{\"nrow\":2,\"ncol\":2,\"columns\":[\"year\",\"letter\"],\"kotlin_dataframe\":[{\"year\":1998,\"letter\":{\"B\":[{\"letter\":\"B\",\"num\":0.6246292191371761,\"date\":\"1998-01-02\"}],\"C\":[{\"letter\":\"C\",\"num\":1.126393826325953,\"date\":\"1998-06-22\"},{\"letter\":\"C\",\"num\":1.4376621993807677,\"date\":\"1998-12-25\"}],\"A\":[{\"letter\":\"A\",\"num\":1.9917663922907851,\"date\":\"1998-01-20\"}]}},{\"year\":1999,\"letter\":{\"B\":[{\"letter\":\"B\",\"num\":0.6762208162903859,\"date\":\"1999-03-12\"},{\"letter\":\"B\",\"num\":1.4920787730184157,\"date\":\"1999-12-24\"},{\"letter\":\"B\",\"num\":-0.07756122563035872,\"date\":\"1999-06-24\"}],\"C\":[{\"letter\":\"C\",\"num\":-1.5985374620272976,\"date\":\"1999-02-09\"}],\"A\":[{\"letter\":\"A\",\"num\":-0.8581918080882499,\"date\":\"1999-11-21\"},{\"letter\":\"A\",\"num\":0.27096900216864345,\"date\":\"1999-12-20\"}]}}]}",
      "text/html": [
       "        <html>\n",
       "        <head>\n",
       "            <style type=\"text/css\">\n",
       "                :root {\n",
       "    --background: #fff;\n",
       "    --background-odd: #f5f5f5;\n",
       "    --background-hover: #d9edfd;\n",
       "    --header-text-color: #474747;\n",
       "    --text-color: #848484;\n",
       "    --text-color-dark: #000;\n",
       "    --text-color-medium: #737373;\n",
       "    --text-color-pale: #b3b3b3;\n",
       "    --inner-border-color: #aaa;\n",
       "    --bold-border-color: #000;\n",
       "    --link-color: #296eaa;\n",
       "    --link-color-pale: #296eaa;\n",
       "    --link-hover: #1a466c;\n",
       "}\n",
       "\n",
       ":root[theme=\"dark\"], :root [data-jp-theme-light=\"false\"], .dataframe_dark{\n",
       "    --background: #303030;\n",
       "    --background-odd: #3c3c3c;\n",
       "    --background-hover: #464646;\n",
       "    --header-text-color: #dddddd;\n",
       "    --text-color: #b3b3b3;\n",
       "    --text-color-dark: #dddddd;\n",
       "    --text-color-medium: #b2b2b2;\n",
       "    --text-color-pale: #737373;\n",
       "    --inner-border-color: #707070;\n",
       "    --bold-border-color: #777777;\n",
       "    --link-color: #008dc0;\n",
       "    --link-color-pale: #97e1fb;\n",
       "    --link-hover: #00688e;\n",
       "}\n",
       "\n",
       "p.dataframe_description {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "table.dataframe {\n",
       "    font-family: \"Helvetica Neue\", Helvetica, Arial, sans-serif;\n",
       "    font-size: 12px;\n",
       "    background-color: var(--background);\n",
       "    color: var(--text-color-dark);\n",
       "    border: none;\n",
       "    border-collapse: collapse;\n",
       "}\n",
       "\n",
       "table.dataframe th, td {\n",
       "    padding: 6px;\n",
       "    border: 1px solid transparent;\n",
       "    text-align: left;\n",
       "}\n",
       "\n",
       "table.dataframe th {\n",
       "    background-color: var(--background);\n",
       "    color: var(--header-text-color);\n",
       "}\n",
       "\n",
       "table.dataframe td {\n",
       "    vertical-align: top;\n",
       "}\n",
       "\n",
       "table.dataframe th.bottomBorder {\n",
       "    border-bottom-color: var(--bold-border-color);\n",
       "}\n",
       "\n",
       "table.dataframe tbody > tr:nth-child(odd) {\n",
       "    background: var(--background-odd);\n",
       "}\n",
       "\n",
       "table.dataframe tbody > tr:nth-child(even) {\n",
       "    background: var(--background);\n",
       "}\n",
       "\n",
       "table.dataframe tbody > tr:hover {\n",
       "    background: var(--background-hover);\n",
       "}\n",
       "\n",
       "table.dataframe a {\n",
       "    cursor: pointer;\n",
       "    color: var(--link-color);\n",
       "    text-decoration: none;\n",
       "}\n",
       "\n",
       "table.dataframe tr:hover > td a {\n",
       "    color: var(--link-color-pale);\n",
       "}\n",
       "\n",
       "table.dataframe a:hover {\n",
       "    color: var(--link-hover);\n",
       "    text-decoration: underline;\n",
       "}\n",
       "\n",
       "table.dataframe img {\n",
       "    max-width: fit-content;\n",
       "}\n",
       "\n",
       "table.dataframe th.complex {\n",
       "    background-color: var(--background);\n",
       "    border: 1px solid var(--background);\n",
       "}\n",
       "\n",
       "table.dataframe .leftBorder {\n",
       "    border-left-color: var(--inner-border-color);\n",
       "}\n",
       "\n",
       "table.dataframe .rightBorder {\n",
       "    border-right-color: var(--inner-border-color);\n",
       "}\n",
       "\n",
       "table.dataframe .rightAlign {\n",
       "    text-align: right;\n",
       "}\n",
       "\n",
       "table.dataframe .expanderSvg {\n",
       "    width: 8px;\n",
       "    height: 8px;\n",
       "    margin-right: 3px;\n",
       "}\n",
       "\n",
       "table.dataframe .expander {\n",
       "    display: flex;\n",
       "    align-items: center;\n",
       "}\n",
       "\n",
       "/* formatting */\n",
       "\n",
       "table.dataframe .null {\n",
       "    color: var(--text-color-pale);\n",
       "}\n",
       "\n",
       "table.dataframe .structural {\n",
       "    color: var(--text-color-medium);\n",
       "    font-weight: bold;\n",
       "}\n",
       "\n",
       "table.dataframe .dataFrameCaption {\n",
       "    font-weight: bold;\n",
       "}\n",
       "\n",
       "table.dataframe .numbers {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "table.dataframe td:hover .formatted .structural, .null {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "table.dataframe tr:hover .formatted .structural, .null {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "\n",
       "\n",
       "\n",
       "            </style>\n",
       "        </head>\n",
       "        <body>\n",
       "            \n",
       "<table class=\"dataframe\" id=\"df_-2046820334\"></table>\n",
       "\n",
       "<p class=\"dataframe_description\">PivotGroupBy</p>\n",
       "        </body>\n",
       "        <script>\n",
       "            \n",
       "/*<!--*/\n",
       "call_DataFrame(function() { DataFrame.addTable({ cols: [{ name: \"<span title=\\\"year: Int\\\">year</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1998</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1999</span></span>\"] }, \n",
       "{ name: \"<span title=\\\"B: DataFrame\\\">B</span>\", children: [], rightAlign: false, values: [{ frameId: -2046820333, value: \"<b>DataFrame 1 x 3</b>\" },{ frameId: -2046820332, value: \"<b>DataFrame 3 x 3</b>\" }] }, \n",
       "{ name: \"<span title=\\\"C: DataFrame\\\">C</span>\", children: [], rightAlign: false, values: [{ frameId: -2046820331, value: \"<b>DataFrame 2 x 3</b>\" },{ frameId: -2046820330, value: \"<b>DataFrame 1 x 3</b>\" }] }, \n",
       "{ name: \"<span title=\\\"A: DataFrame\\\">A</span>\", children: [], rightAlign: false, values: [{ frameId: -2046820329, value: \"<b>DataFrame 1 x 3</b>\" },{ frameId: -2046820328, value: \"<b>DataFrame 2 x 3</b>\" }] }, \n",
       "{ name: \"<span title=\\\"letter: DataRow\\\">letter</span>\", children: [1, 2, 3], rightAlign: false, values: [\"<span class=\\\"formatted\\\" title=\\\"B:    letter      num       date\\n 0      B 0.624629 1998-01-02\\n\\nC:    letter      num       date\\n 0      C 1.126394 1998-06-22\\n 1      C 1.437662 1998-12-25\\n\\nA:    letter      num       date\\n 0      A 1.991766 1998-01-20\\n\\\"><span class=\\\"structural\\\">{ </span><span class=\\\"structural\\\">B: </span><span class=\\\"dataFrameCaption\\\">DataFrame [1 x 3]</span><span class=\\\"structural\\\">, </span><span class=\\\"structural\\\">C: </span><span class=\\\"dataFrameCaption\\\">Da<span class=\\\"structural\\\">...</span></span><span class=\\\"structural\\\">, </span><span class=\\\"structural\\\">A...</span><span class=\\\"structural\\\"> }</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"B:    letter       num       date\\n 0      B  0.676221 1999-03-12\\n 1      B  1.492079 1999-12-24\\n 2      B -0.077561 1999-06-24\\n\\nC:    letter       num       date\\n 0      C -1.598537 1999-02-09\\n\\nA:    letter       num       date\\n 0      A -0.858192 1999-11-21\\n 1      A  0.270969 1999-12-20\\n\\\"><span class=\\\"structural\\\">{ </span><span class=\\\"structural\\\">B: </span><span class=\\\"dataFrameCaption\\\">DataFrame [3 x 3]</span><span class=\\\"structural\\\">, </span><span class=\\\"structural\\\">C: </span><span class=\\\"dataFrameCaption\\\">Da<span class=\\\"structural\\\">...</span></span><span class=\\\"structural\\\">, </span><span class=\\\"structural\\\">A...</span><span class=\\\"structural\\\"> }</span></span>\"] }, \n",
       "], id: -2046820334, rootId: -2046820334, totalRows: 2 } ) });\n",
       "/*-->*/\n",
       "\n",
       "/*<!--*/\n",
       "call_DataFrame(function() { DataFrame.addTable({ cols: [{ name: \"<span title=\\\"letter: String\\\">letter</span>\", children: [], rightAlign: false, values: [\"B\"] }, \n",
       "{ name: \"<span title=\\\"num: Double\\\">num</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">0.624629</span></span>\"] }, \n",
       "{ name: \"<span title=\\\"date: kotlinx.datetime.LocalDate\\\">date</span>\", children: [], rightAlign: false, values: [\"1998-01-02\"] }, \n",
       "], id: -2046820333, rootId: -2046820334, totalRows: 1 } ) });\n",
       "/*-->*/\n",
       "\n",
       "/*<!--*/\n",
       "call_DataFrame(function() { DataFrame.addTable({ cols: [{ name: \"<span title=\\\"letter: String\\\">letter</span>\", children: [], rightAlign: false, values: [\"B\",\"B\",\"B\"] }, \n",
       "{ name: \"<span title=\\\"num: Double\\\">num</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">0.676221</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1.492079</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">-0.077561</span></span>\"] }, \n",
       "{ name: \"<span title=\\\"date: kotlinx.datetime.LocalDate\\\">date</span>\", children: [], rightAlign: false, values: [\"1999-03-12\",\"1999-12-24\",\"1999-06-24\"] }, \n",
       "], id: -2046820332, rootId: -2046820334, totalRows: 3 } ) });\n",
       "/*-->*/\n",
       "\n",
       "/*<!--*/\n",
       "call_DataFrame(function() { DataFrame.addTable({ cols: [{ name: \"<span title=\\\"letter: String\\\">letter</span>\", children: [], rightAlign: false, values: [\"C\",\"C\"] }, \n",
       "{ name: \"<span title=\\\"num: Double\\\">num</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1.126394</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1.437662</span></span>\"] }, \n",
       "{ name: \"<span title=\\\"date: kotlinx.datetime.LocalDate\\\">date</span>\", children: [], rightAlign: false, values: [\"1998-06-22\",\"1998-12-25\"] }, \n",
       "], id: -2046820331, rootId: -2046820334, totalRows: 2 } ) });\n",
       "/*-->*/\n",
       "\n",
       "/*<!--*/\n",
       "call_DataFrame(function() { DataFrame.addTable({ cols: [{ name: \"<span title=\\\"letter: String\\\">letter</span>\", children: [], rightAlign: false, values: [\"C\"] }, \n",
       "{ name: \"<span title=\\\"num: Double\\\">num</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">-1.598537</span></span>\"] }, \n",
       "{ name: \"<span title=\\\"date: kotlinx.datetime.LocalDate\\\">date</span>\", children: [], rightAlign: false, values: [\"1999-02-09\"] }, \n",
       "], id: -2046820330, rootId: -2046820334, totalRows: 1 } ) });\n",
       "/*-->*/\n",
       "\n",
       "/*<!--*/\n",
       "call_DataFrame(function() { DataFrame.addTable({ cols: [{ name: \"<span title=\\\"letter: String\\\">letter</span>\", children: [], rightAlign: false, values: [\"A\"] }, \n",
       "{ name: \"<span title=\\\"num: Double\\\">num</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1.991766</span></span>\"] }, \n",
       "{ name: \"<span title=\\\"date: kotlinx.datetime.LocalDate\\\">date</span>\", children: [], rightAlign: false, values: [\"1998-01-20\"] }, \n",
       "], id: -2046820329, rootId: -2046820334, totalRows: 1 } ) });\n",
       "/*-->*/\n",
       "\n",
       "/*<!--*/\n",
       "call_DataFrame(function() { DataFrame.addTable({ cols: [{ name: \"<span title=\\\"letter: String\\\">letter</span>\", children: [], rightAlign: false, values: [\"A\",\"A\"] }, \n",
       "{ name: \"<span title=\\\"num: Double\\\">num</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">-0.858192</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">0.270969</span></span>\"] }, \n",
       "{ name: \"<span title=\\\"date: kotlinx.datetime.LocalDate\\\">date</span>\", children: [], rightAlign: false, values: [\"1999-11-21\",\"1999-12-20\"] }, \n",
       "], id: -2046820328, rootId: -2046820334, totalRows: 2 } ) });\n",
       "/*-->*/\n",
       "\n",
       "call_DataFrame(function() { DataFrame.renderTable(-2046820334) });\n",
       "\n",
       "\n",
       "        </script>\n",
       "        </html>"
      ]
     },
     "execution_count": 8,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "dfd.groupBy { date.map { it.year } named \"year\" }.pivot { letter }"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "eb19c608-4b0f-472d-89c1-4450edaf19a8",
   "metadata": {},
   "source": [
    "The result we have computed are not directly usable: as said before, the split-apply-combine strategy require that grouped data are aggregated and reduced with an application of a combining operation."
   ]
  },
  {
   "cell_type": "markdown",
   "id": "ce6641d0-d24b-4436-b3d2-26a45d6fcdb8",
   "metadata": {},
   "source": [
    "## Aggregation (Apply and Combine)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "0b6c51e8-eb88-4a01-aeca-6b91fe71c387",
   "metadata": {},
   "source": [
    "With grouped data, we can compute a large variety of operations on groups. The most common operations are defined in Kotlin DataFrame library, like `mean()`, `sum()` or `count()`. \n",
    "\n",
    "For example, if we take the dataframe we grouped above, we can compute the mean of the values inside each group:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 9,
   "id": "63251f86-a72d-4269-9696-bbca9a312d6b",
   "metadata": {
    "tags": []
   },
   "outputs": [
    {
     "data": {
      "application/kotlindataframe+json": "{\"nrow\":2,\"ncol\":2,\"columns\":[\"year\",\"letter\"],\"kotlin_dataframe\":[{\"year\":1998,\"letter\":{\"B\":0.6246292191371761,\"C\":1.2820280128533603,\"A\":1.9917663922907851}},{\"year\":1999,\"letter\":{\"B\":0.6969127878928143,\"C\":-1.5985374620272976,\"A\":-0.29361140295980326}}]}",
      "text/html": [
       "        <html>\n",
       "        <head>\n",
       "            <style type=\"text/css\">\n",
       "                :root {\n",
       "    --background: #fff;\n",
       "    --background-odd: #f5f5f5;\n",
       "    --background-hover: #d9edfd;\n",
       "    --header-text-color: #474747;\n",
       "    --text-color: #848484;\n",
       "    --text-color-dark: #000;\n",
       "    --text-color-medium: #737373;\n",
       "    --text-color-pale: #b3b3b3;\n",
       "    --inner-border-color: #aaa;\n",
       "    --bold-border-color: #000;\n",
       "    --link-color: #296eaa;\n",
       "    --link-color-pale: #296eaa;\n",
       "    --link-hover: #1a466c;\n",
       "}\n",
       "\n",
       ":root[theme=\"dark\"], :root [data-jp-theme-light=\"false\"], .dataframe_dark{\n",
       "    --background: #303030;\n",
       "    --background-odd: #3c3c3c;\n",
       "    --background-hover: #464646;\n",
       "    --header-text-color: #dddddd;\n",
       "    --text-color: #b3b3b3;\n",
       "    --text-color-dark: #dddddd;\n",
       "    --text-color-medium: #b2b2b2;\n",
       "    --text-color-pale: #737373;\n",
       "    --inner-border-color: #707070;\n",
       "    --bold-border-color: #777777;\n",
       "    --link-color: #008dc0;\n",
       "    --link-color-pale: #97e1fb;\n",
       "    --link-hover: #00688e;\n",
       "}\n",
       "\n",
       "p.dataframe_description {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "table.dataframe {\n",
       "    font-family: \"Helvetica Neue\", Helvetica, Arial, sans-serif;\n",
       "    font-size: 12px;\n",
       "    background-color: var(--background);\n",
       "    color: var(--text-color-dark);\n",
       "    border: none;\n",
       "    border-collapse: collapse;\n",
       "}\n",
       "\n",
       "table.dataframe th, td {\n",
       "    padding: 6px;\n",
       "    border: 1px solid transparent;\n",
       "    text-align: left;\n",
       "}\n",
       "\n",
       "table.dataframe th {\n",
       "    background-color: var(--background);\n",
       "    color: var(--header-text-color);\n",
       "}\n",
       "\n",
       "table.dataframe td {\n",
       "    vertical-align: top;\n",
       "}\n",
       "\n",
       "table.dataframe th.bottomBorder {\n",
       "    border-bottom-color: var(--bold-border-color);\n",
       "}\n",
       "\n",
       "table.dataframe tbody > tr:nth-child(odd) {\n",
       "    background: var(--background-odd);\n",
       "}\n",
       "\n",
       "table.dataframe tbody > tr:nth-child(even) {\n",
       "    background: var(--background);\n",
       "}\n",
       "\n",
       "table.dataframe tbody > tr:hover {\n",
       "    background: var(--background-hover);\n",
       "}\n",
       "\n",
       "table.dataframe a {\n",
       "    cursor: pointer;\n",
       "    color: var(--link-color);\n",
       "    text-decoration: none;\n",
       "}\n",
       "\n",
       "table.dataframe tr:hover > td a {\n",
       "    color: var(--link-color-pale);\n",
       "}\n",
       "\n",
       "table.dataframe a:hover {\n",
       "    color: var(--link-hover);\n",
       "    text-decoration: underline;\n",
       "}\n",
       "\n",
       "table.dataframe img {\n",
       "    max-width: fit-content;\n",
       "}\n",
       "\n",
       "table.dataframe th.complex {\n",
       "    background-color: var(--background);\n",
       "    border: 1px solid var(--background);\n",
       "}\n",
       "\n",
       "table.dataframe .leftBorder {\n",
       "    border-left-color: var(--inner-border-color);\n",
       "}\n",
       "\n",
       "table.dataframe .rightBorder {\n",
       "    border-right-color: var(--inner-border-color);\n",
       "}\n",
       "\n",
       "table.dataframe .rightAlign {\n",
       "    text-align: right;\n",
       "}\n",
       "\n",
       "table.dataframe .expanderSvg {\n",
       "    width: 8px;\n",
       "    height: 8px;\n",
       "    margin-right: 3px;\n",
       "}\n",
       "\n",
       "table.dataframe .expander {\n",
       "    display: flex;\n",
       "    align-items: center;\n",
       "}\n",
       "\n",
       "/* formatting */\n",
       "\n",
       "table.dataframe .null {\n",
       "    color: var(--text-color-pale);\n",
       "}\n",
       "\n",
       "table.dataframe .structural {\n",
       "    color: var(--text-color-medium);\n",
       "    font-weight: bold;\n",
       "}\n",
       "\n",
       "table.dataframe .dataFrameCaption {\n",
       "    font-weight: bold;\n",
       "}\n",
       "\n",
       "table.dataframe .numbers {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "table.dataframe td:hover .formatted .structural, .null {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "table.dataframe tr:hover .formatted .structural, .null {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "\n",
       "\n",
       "\n",
       "            </style>\n",
       "        </head>\n",
       "        <body>\n",
       "            \n",
       "<table class=\"dataframe\" id=\"df_-2046820327\"></table>\n",
       "\n",
       "<p class=\"dataframe_description\">DataFrame: rowsCount = 2, columnsCount = 2</p>\n",
       "        </body>\n",
       "        <script>\n",
       "            \n",
       "/*<!--*/\n",
       "call_DataFrame(function() { DataFrame.addTable({ cols: [{ name: \"<span title=\\\"year: Int\\\">year</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1998</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1999</span></span>\"] }, \n",
       "{ name: \"<span title=\\\"B: Double\\\">B</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">0.624629</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">0.696913</span></span>\"] }, \n",
       "{ name: \"<span title=\\\"C: Double\\\">C</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1.282028</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">-1.598537</span></span>\"] }, \n",
       "{ name: \"<span title=\\\"A: Double\\\">A</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1.991766</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">-0.293611</span></span>\"] }, \n",
       "{ name: \"<span title=\\\"letter: DataRow\\\">letter</span>\", children: [1, 2, 3], rightAlign: false, values: [\"<span class=\\\"formatted\\\" title=\\\"B: 0.6246292191371761\\nC: 1.2820280128533603\\nA: 1.9917663922907851\\\"><span class=\\\"structural\\\">{ </span><span class=\\\"structural\\\">B: </span><span class=\\\"numbers\\\">0.6</span><span class=\\\"structural\\\">, </span><span class=\\\"structural\\\">C: </span><span class=\\\"numbers\\\">1.3</span><span class=\\\"structural\\\">, </span><span class=\\\"structural\\\">A: </span><span class=\\\"numbers\\\">2.0</span><span class=\\\"structural\\\"> }</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"B: 0.6969127878928143\\nC: -1.5985374620272976\\nA: -0.29361140295980326\\\"><span class=\\\"structural\\\">{ </span><span class=\\\"structural\\\">B: </span><span class=\\\"numbers\\\">0.7</span><span class=\\\"structural\\\">, </span><span class=\\\"structural\\\">C: </span><span class=\\\"numbers\\\">-1.6</span><span class=\\\"structural\\\">, </span><span class=\\\"structural\\\">A: </span><span class=\\\"numbers\\\">-0.3</span><span class=\\\"structural\\\"> }</span></span>\"] }, \n",
       "], id: -2046820327, rootId: -2046820327, totalRows: 2 } ) });\n",
       "/*-->*/\n",
       "\n",
       "call_DataFrame(function() { DataFrame.renderTable(-2046820327) });\n",
       "\n",
       "\n",
       "        </script>\n",
       "        </html>"
      ]
     },
     "execution_count": 9,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "val groupedDf = dfd.groupBy { date.map { it.year } named \"year\" }\n",
    "    .pivot { letter }\n",
    "\n",
    "groupedDf.mean()"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "4c8021ba-5058-4bac-968f-3f54d94d19c6",
   "metadata": {},
   "source": [
    "The same exact result can be archived in python with a very similar instruction:\n",
    "```python\n",
    "dfd.groupby([df['date'].dt.year, 'letter'])\\\n",
    "    .mean() \\\n",
    "    .unstack('letter')\n",
    "```\n",
    "\n",
    "Note that the `unstack` operation is used to move \"letters\" from the `MultiIndex` object created with \"year\", to columns. It is also possible to call `pivot_table` instead of `unstack`, specifying what will be in place of indices, columns keys and values."
   ]
  },
  {
   "cell_type": "markdown",
   "id": "8825f878-79c7-4f46-886e-d8f2add2e452",
   "metadata": {},
   "source": [
    "Sometimes is useful to flatten the group result, in favor of a better handling of indices. We can use `flatten()`, which affects *every* groups of the dataframe, or `ungroup()` for selecting only one group."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 10,
   "id": "cbeb9081-0d54-4ae7-a7b2-41f41445061f",
   "metadata": {
    "tags": []
   },
   "outputs": [
    {
     "data": {
      "application/kotlindataframe+json": "{\"nrow\":2,\"ncol\":4,\"columns\":[\"year\",\"B\",\"C\",\"A\"],\"kotlin_dataframe\":[{\"year\":1998,\"B\":0.6246292191371761,\"C\":1.2820280128533603,\"A\":1.9917663922907851},{\"year\":1999,\"B\":0.6969127878928143,\"C\":-1.5985374620272976,\"A\":-0.29361140295980326}]}",
      "text/html": [
       "        <html>\n",
       "        <head>\n",
       "            <style type=\"text/css\">\n",
       "                :root {\n",
       "    --background: #fff;\n",
       "    --background-odd: #f5f5f5;\n",
       "    --background-hover: #d9edfd;\n",
       "    --header-text-color: #474747;\n",
       "    --text-color: #848484;\n",
       "    --text-color-dark: #000;\n",
       "    --text-color-medium: #737373;\n",
       "    --text-color-pale: #b3b3b3;\n",
       "    --inner-border-color: #aaa;\n",
       "    --bold-border-color: #000;\n",
       "    --link-color: #296eaa;\n",
       "    --link-color-pale: #296eaa;\n",
       "    --link-hover: #1a466c;\n",
       "}\n",
       "\n",
       ":root[theme=\"dark\"], :root [data-jp-theme-light=\"false\"], .dataframe_dark{\n",
       "    --background: #303030;\n",
       "    --background-odd: #3c3c3c;\n",
       "    --background-hover: #464646;\n",
       "    --header-text-color: #dddddd;\n",
       "    --text-color: #b3b3b3;\n",
       "    --text-color-dark: #dddddd;\n",
       "    --text-color-medium: #b2b2b2;\n",
       "    --text-color-pale: #737373;\n",
       "    --inner-border-color: #707070;\n",
       "    --bold-border-color: #777777;\n",
       "    --link-color: #008dc0;\n",
       "    --link-color-pale: #97e1fb;\n",
       "    --link-hover: #00688e;\n",
       "}\n",
       "\n",
       "p.dataframe_description {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "table.dataframe {\n",
       "    font-family: \"Helvetica Neue\", Helvetica, Arial, sans-serif;\n",
       "    font-size: 12px;\n",
       "    background-color: var(--background);\n",
       "    color: var(--text-color-dark);\n",
       "    border: none;\n",
       "    border-collapse: collapse;\n",
       "}\n",
       "\n",
       "table.dataframe th, td {\n",
       "    padding: 6px;\n",
       "    border: 1px solid transparent;\n",
       "    text-align: left;\n",
       "}\n",
       "\n",
       "table.dataframe th {\n",
       "    background-color: var(--background);\n",
       "    color: var(--header-text-color);\n",
       "}\n",
       "\n",
       "table.dataframe td {\n",
       "    vertical-align: top;\n",
       "}\n",
       "\n",
       "table.dataframe th.bottomBorder {\n",
       "    border-bottom-color: var(--bold-border-color);\n",
       "}\n",
       "\n",
       "table.dataframe tbody > tr:nth-child(odd) {\n",
       "    background: var(--background-odd);\n",
       "}\n",
       "\n",
       "table.dataframe tbody > tr:nth-child(even) {\n",
       "    background: var(--background);\n",
       "}\n",
       "\n",
       "table.dataframe tbody > tr:hover {\n",
       "    background: var(--background-hover);\n",
       "}\n",
       "\n",
       "table.dataframe a {\n",
       "    cursor: pointer;\n",
       "    color: var(--link-color);\n",
       "    text-decoration: none;\n",
       "}\n",
       "\n",
       "table.dataframe tr:hover > td a {\n",
       "    color: var(--link-color-pale);\n",
       "}\n",
       "\n",
       "table.dataframe a:hover {\n",
       "    color: var(--link-hover);\n",
       "    text-decoration: underline;\n",
       "}\n",
       "\n",
       "table.dataframe img {\n",
       "    max-width: fit-content;\n",
       "}\n",
       "\n",
       "table.dataframe th.complex {\n",
       "    background-color: var(--background);\n",
       "    border: 1px solid var(--background);\n",
       "}\n",
       "\n",
       "table.dataframe .leftBorder {\n",
       "    border-left-color: var(--inner-border-color);\n",
       "}\n",
       "\n",
       "table.dataframe .rightBorder {\n",
       "    border-right-color: var(--inner-border-color);\n",
       "}\n",
       "\n",
       "table.dataframe .rightAlign {\n",
       "    text-align: right;\n",
       "}\n",
       "\n",
       "table.dataframe .expanderSvg {\n",
       "    width: 8px;\n",
       "    height: 8px;\n",
       "    margin-right: 3px;\n",
       "}\n",
       "\n",
       "table.dataframe .expander {\n",
       "    display: flex;\n",
       "    align-items: center;\n",
       "}\n",
       "\n",
       "/* formatting */\n",
       "\n",
       "table.dataframe .null {\n",
       "    color: var(--text-color-pale);\n",
       "}\n",
       "\n",
       "table.dataframe .structural {\n",
       "    color: var(--text-color-medium);\n",
       "    font-weight: bold;\n",
       "}\n",
       "\n",
       "table.dataframe .dataFrameCaption {\n",
       "    font-weight: bold;\n",
       "}\n",
       "\n",
       "table.dataframe .numbers {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "table.dataframe td:hover .formatted .structural, .null {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "table.dataframe tr:hover .formatted .structural, .null {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "\n",
       "\n",
       "\n",
       "            </style>\n",
       "        </head>\n",
       "        <body>\n",
       "            \n",
       "<table class=\"dataframe\" id=\"df_-2046820326\"></table>\n",
       "\n",
       "<p class=\"dataframe_description\">DataFrame: rowsCount = 2, columnsCount = 4</p>\n",
       "        </body>\n",
       "        <script>\n",
       "            \n",
       "/*<!--*/\n",
       "call_DataFrame(function() { DataFrame.addTable({ cols: [{ name: \"<span title=\\\"year: Int\\\">year</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1998</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1999</span></span>\"] }, \n",
       "{ name: \"<span title=\\\"B: Double\\\">B</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">0.624629</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">0.696913</span></span>\"] }, \n",
       "{ name: \"<span title=\\\"C: Double\\\">C</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1.282028</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">-1.598537</span></span>\"] }, \n",
       "{ name: \"<span title=\\\"A: Double\\\">A</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1.991766</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">-0.293611</span></span>\"] }, \n",
       "], id: -2046820326, rootId: -2046820326, totalRows: 2 } ) });\n",
       "/*-->*/\n",
       "\n",
       "call_DataFrame(function() { DataFrame.renderTable(-2046820326) });\n",
       "\n",
       "\n",
       "        </script>\n",
       "        </html>"
      ]
     },
     "execution_count": 10,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "groupedDf.mean().ungroup(\"letter\")"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "c994f808-322d-445f-9697-a9f9a61c5207",
   "metadata": {},
   "source": [
    "Every time that we create a `GroupBy` or `Pivot` object, and after applying the needed transformations, we can aggregate or reduce all data with several functions all computed in the same time, inside the `aggregate` function. Each operation can then be mapped in the corresponding column. \n",
    "\n",
    "If we want to compute the sum, the mean and the number of rows of each group, we can write:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 11,
   "id": "8280a3ba-ef40-4410-af5d-736c371cca92",
   "metadata": {
    "tags": []
   },
   "outputs": [
    {
     "data": {
      "application/kotlindataframe+json": "{\"nrow\":2,\"ncol\":4,\"columns\":[\"year\",\"B\",\"C\",\"A\"],\"kotlin_dataframe\":[{\"year\":1998,\"B\":{\"sum\":0.6246292191371761,\"mean\":0.6246292191371761,\"count\":1},\"C\":{\"sum\":2.5640560257067206,\"mean\":1.2820280128533603,\"count\":2},\"A\":{\"sum\":1.9917663922907851,\"mean\":1.9917663922907851,\"count\":1}},{\"year\":1999,\"B\":{\"sum\":2.090738363678443,\"mean\":0.6969127878928143,\"count\":3},\"C\":{\"sum\":-1.5985374620272976,\"mean\":-1.5985374620272976,\"count\":1},\"A\":{\"sum\":-0.5872228059196065,\"mean\":-0.29361140295980326,\"count\":2}}]}",
      "text/html": [
       "        <html>\n",
       "        <head>\n",
       "            <style type=\"text/css\">\n",
       "                :root {\n",
       "    --background: #fff;\n",
       "    --background-odd: #f5f5f5;\n",
       "    --background-hover: #d9edfd;\n",
       "    --header-text-color: #474747;\n",
       "    --text-color: #848484;\n",
       "    --text-color-dark: #000;\n",
       "    --text-color-medium: #737373;\n",
       "    --text-color-pale: #b3b3b3;\n",
       "    --inner-border-color: #aaa;\n",
       "    --bold-border-color: #000;\n",
       "    --link-color: #296eaa;\n",
       "    --link-color-pale: #296eaa;\n",
       "    --link-hover: #1a466c;\n",
       "}\n",
       "\n",
       ":root[theme=\"dark\"], :root [data-jp-theme-light=\"false\"], .dataframe_dark{\n",
       "    --background: #303030;\n",
       "    --background-odd: #3c3c3c;\n",
       "    --background-hover: #464646;\n",
       "    --header-text-color: #dddddd;\n",
       "    --text-color: #b3b3b3;\n",
       "    --text-color-dark: #dddddd;\n",
       "    --text-color-medium: #b2b2b2;\n",
       "    --text-color-pale: #737373;\n",
       "    --inner-border-color: #707070;\n",
       "    --bold-border-color: #777777;\n",
       "    --link-color: #008dc0;\n",
       "    --link-color-pale: #97e1fb;\n",
       "    --link-hover: #00688e;\n",
       "}\n",
       "\n",
       "p.dataframe_description {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "table.dataframe {\n",
       "    font-family: \"Helvetica Neue\", Helvetica, Arial, sans-serif;\n",
       "    font-size: 12px;\n",
       "    background-color: var(--background);\n",
       "    color: var(--text-color-dark);\n",
       "    border: none;\n",
       "    border-collapse: collapse;\n",
       "}\n",
       "\n",
       "table.dataframe th, td {\n",
       "    padding: 6px;\n",
       "    border: 1px solid transparent;\n",
       "    text-align: left;\n",
       "}\n",
       "\n",
       "table.dataframe th {\n",
       "    background-color: var(--background);\n",
       "    color: var(--header-text-color);\n",
       "}\n",
       "\n",
       "table.dataframe td {\n",
       "    vertical-align: top;\n",
       "}\n",
       "\n",
       "table.dataframe th.bottomBorder {\n",
       "    border-bottom-color: var(--bold-border-color);\n",
       "}\n",
       "\n",
       "table.dataframe tbody > tr:nth-child(odd) {\n",
       "    background: var(--background-odd);\n",
       "}\n",
       "\n",
       "table.dataframe tbody > tr:nth-child(even) {\n",
       "    background: var(--background);\n",
       "}\n",
       "\n",
       "table.dataframe tbody > tr:hover {\n",
       "    background: var(--background-hover);\n",
       "}\n",
       "\n",
       "table.dataframe a {\n",
       "    cursor: pointer;\n",
       "    color: var(--link-color);\n",
       "    text-decoration: none;\n",
       "}\n",
       "\n",
       "table.dataframe tr:hover > td a {\n",
       "    color: var(--link-color-pale);\n",
       "}\n",
       "\n",
       "table.dataframe a:hover {\n",
       "    color: var(--link-hover);\n",
       "    text-decoration: underline;\n",
       "}\n",
       "\n",
       "table.dataframe img {\n",
       "    max-width: fit-content;\n",
       "}\n",
       "\n",
       "table.dataframe th.complex {\n",
       "    background-color: var(--background);\n",
       "    border: 1px solid var(--background);\n",
       "}\n",
       "\n",
       "table.dataframe .leftBorder {\n",
       "    border-left-color: var(--inner-border-color);\n",
       "}\n",
       "\n",
       "table.dataframe .rightBorder {\n",
       "    border-right-color: var(--inner-border-color);\n",
       "}\n",
       "\n",
       "table.dataframe .rightAlign {\n",
       "    text-align: right;\n",
       "}\n",
       "\n",
       "table.dataframe .expanderSvg {\n",
       "    width: 8px;\n",
       "    height: 8px;\n",
       "    margin-right: 3px;\n",
       "}\n",
       "\n",
       "table.dataframe .expander {\n",
       "    display: flex;\n",
       "    align-items: center;\n",
       "}\n",
       "\n",
       "/* formatting */\n",
       "\n",
       "table.dataframe .null {\n",
       "    color: var(--text-color-pale);\n",
       "}\n",
       "\n",
       "table.dataframe .structural {\n",
       "    color: var(--text-color-medium);\n",
       "    font-weight: bold;\n",
       "}\n",
       "\n",
       "table.dataframe .dataFrameCaption {\n",
       "    font-weight: bold;\n",
       "}\n",
       "\n",
       "table.dataframe .numbers {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "table.dataframe td:hover .formatted .structural, .null {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "table.dataframe tr:hover .formatted .structural, .null {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "\n",
       "\n",
       "\n",
       "            </style>\n",
       "        </head>\n",
       "        <body>\n",
       "            \n",
       "<table class=\"dataframe\" id=\"df_-2046820325\"></table>\n",
       "\n",
       "<p class=\"dataframe_description\">DataFrame: rowsCount = 2, columnsCount = 4</p>\n",
       "        </body>\n",
       "        <script>\n",
       "            \n",
       "/*<!--*/\n",
       "call_DataFrame(function() { DataFrame.addTable({ cols: [{ name: \"<span title=\\\"year: Int\\\">year</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1998</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1999</span></span>\"] }, \n",
       "{ name: \"<span title=\\\"sum: Double\\\">sum</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">0.624629</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">2.090738</span></span>\"] }, \n",
       "{ name: \"<span title=\\\"mean: Double\\\">mean</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">0.624629</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">0.696913</span></span>\"] }, \n",
       "{ name: \"<span title=\\\"count: Int\\\">count</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">3</span></span>\"] }, \n",
       "{ name: \"<span title=\\\"B: DataRow\\\">B</span>\", children: [1, 2, 3], rightAlign: false, values: [\"<span class=\\\"formatted\\\" title=\\\"sum: 0.6246292191371761\\nmean: 0.6246292191371761\\ncount: 1\\\"><span class=\\\"structural\\\">{ </span><span class=\\\"structural\\\">sum: </span><span class=\\\"numbers\\\">0.6</span><span class=\\\"structural\\\">, </span><span class=\\\"structural\\\">mean: </span><span class=\\\"numbers\\\">0.6</span><span class=\\\"structural\\\">, </span><span class=\\\"structural\\\">count: </span><span class=\\\"numbers\\\">1</span><span class=\\\"structural\\\"> }</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"sum: 2.090738363678443\\nmean: 0.6969127878928143\\ncount: 3\\\"><span class=\\\"structural\\\">{ </span><span class=\\\"structural\\\">sum: </span><span class=\\\"numbers\\\">2.1</span><span class=\\\"structural\\\">, </span><span class=\\\"structural\\\">mean: </span><span class=\\\"numbers\\\">0.7</span><span class=\\\"structural\\\">, </span><span class=\\\"structural\\\">count: </span><span class=\\\"numbers\\\">3</span><span class=\\\"structural\\\"> }</span></span>\"] }, \n",
       "{ name: \"<span title=\\\"sum: Double\\\">sum</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">2.564056</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">-1.598537</span></span>\"] }, \n",
       "{ name: \"<span title=\\\"mean: Double\\\">mean</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1.282028</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">-1.598537</span></span>\"] }, \n",
       "{ name: \"<span title=\\\"count: Int\\\">count</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">2</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1</span></span>\"] }, \n",
       "{ name: \"<span title=\\\"C: DataRow\\\">C</span>\", children: [5, 6, 7], rightAlign: false, values: [\"<span class=\\\"formatted\\\" title=\\\"sum: 2.5640560257067206\\nmean: 1.2820280128533603\\ncount: 2\\\"><span class=\\\"structural\\\">{ </span><span class=\\\"structural\\\">sum: </span><span class=\\\"numbers\\\">2.6</span><span class=\\\"structural\\\">, </span><span class=\\\"structural\\\">mean: </span><span class=\\\"numbers\\\">1.3</span><span class=\\\"structural\\\">, </span><span class=\\\"structural\\\">count: </span><span class=\\\"numbers\\\">2</span><span class=\\\"structural\\\"> }</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"sum: -1.5985374620272976\\nmean: -1.5985374620272976\\ncount: 1\\\"><span class=\\\"structural\\\">{ </span><span class=\\\"structural\\\">sum: </span><span class=\\\"numbers\\\">-1.6</span><span class=\\\"structural\\\">, </span><span class=\\\"structural\\\">mean: </span><span class=\\\"numbers\\\">-1.6</span><span class=\\\"structural\\\">, </span><span class=\\\"structural\\\">count: </span><span class=\\\"numbers\\\">1</span><span class=\\\"structural\\\"> }</span></span>\"] }, \n",
       "{ name: \"<span title=\\\"sum: Double\\\">sum</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1.991766</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">-0.587223</span></span>\"] }, \n",
       "{ name: \"<span title=\\\"mean: Double\\\">mean</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1.991766</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">-0.293611</span></span>\"] }, \n",
       "{ name: \"<span title=\\\"count: Int\\\">count</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">2</span></span>\"] }, \n",
       "{ name: \"<span title=\\\"A: DataRow\\\">A</span>\", children: [9, 10, 11], rightAlign: false, values: [\"<span class=\\\"formatted\\\" title=\\\"sum: 1.9917663922907851\\nmean: 1.9917663922907851\\ncount: 1\\\"><span class=\\\"structural\\\">{ </span><span class=\\\"structural\\\">sum: </span><span class=\\\"numbers\\\">2.0</span><span class=\\\"structural\\\">, </span><span class=\\\"structural\\\">mean: </span><span class=\\\"numbers\\\">2.0</span><span class=\\\"structural\\\">, </span><span class=\\\"structural\\\">count: </span><span class=\\\"numbers\\\">1</span><span class=\\\"structural\\\"> }</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"sum: -0.5872228059196065\\nmean: -0.29361140295980326\\ncount: 2\\\"><span class=\\\"structural\\\">{ </span><span class=\\\"structural\\\">sum: </span><span class=\\\"numbers\\\">-0.6</span><span class=\\\"structural\\\">, </span><span class=\\\"structural\\\">mean: </span><span class=\\\"numbers\\\">-0.3</span><span class=\\\"structural\\\">, </span><span class=\\\"structural\\\">count: </span><span class=\\\"numbers\\\">2</span><span class=\\\"structural\\\"> }</span></span>\"] }, \n",
       "], id: -2046820325, rootId: -2046820325, totalRows: 2 } ) });\n",
       "/*-->*/\n",
       "\n",
       "call_DataFrame(function() { DataFrame.renderTable(-2046820325) });\n",
       "\n",
       "\n",
       "        </script>\n",
       "        </html>"
      ]
     },
     "execution_count": 11,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "groupedDf.aggregate { \n",
    "        sum { num } into \"sum\"\n",
    "        mean { num } into \"mean\"\n",
    "        count() into \"count\"\n",
    "    }.ungroup(\"letter\")"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "63e750f3-37d4-459c-8e0c-f15fdfb2af38",
   "metadata": {},
   "source": [
    "And this approach can be more straightforward than pandas use of `agg` function. If we consider the example above, a similar behavior can be computed in python as follows:\n",
    "```python\n",
    "df.groupby([df['date'].dt.year, 'letter']) \\\n",
    "    .agg([('sum', 'sum'), ('mean', 'mean'), ('count', 'count')]) \\\n",
    "    .unstack('letter')\n",
    "```\n",
    "Where inside `agg` function, we specify the name of the column and the function we want to apply. \n",
    "\n",
    "Both in python and Kotin, inside the aggregation function a custom function can be called. For example let's consider this function:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 12,
   "id": "c971a953-6c5b-4718-9e89-ea7fe22f9210",
   "metadata": {
    "tags": []
   },
   "outputs": [],
   "source": [
    "fun peakToPeak(arr: List<Double>): Double = arr.max() - arr.min()"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "c6e801a9-cef8-434c-932c-31be57416c58",
   "metadata": {},
   "source": [
    "We can the call the function inside the aggregation method:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 13,
   "id": "988a7c20-ea97-43d2-b2fe-55168063b465",
   "metadata": {
    "tags": []
   },
   "outputs": [
    {
     "data": {
      "application/kotlindataframe+json": "{\"nrow\":2,\"ncol\":2,\"columns\":[\"year\",\"letter\"],\"kotlin_dataframe\":[{\"year\":1998,\"letter\":{\"B\":0.0,\"C\":0.3112683730548147,\"A\":0.0}},{\"year\":1999,\"letter\":{\"B\":1.5696399986487743,\"C\":0.0,\"A\":1.1291608102568933}}]}",
      "text/html": [
       "        <html>\n",
       "        <head>\n",
       "            <style type=\"text/css\">\n",
       "                :root {\n",
       "    --background: #fff;\n",
       "    --background-odd: #f5f5f5;\n",
       "    --background-hover: #d9edfd;\n",
       "    --header-text-color: #474747;\n",
       "    --text-color: #848484;\n",
       "    --text-color-dark: #000;\n",
       "    --text-color-medium: #737373;\n",
       "    --text-color-pale: #b3b3b3;\n",
       "    --inner-border-color: #aaa;\n",
       "    --bold-border-color: #000;\n",
       "    --link-color: #296eaa;\n",
       "    --link-color-pale: #296eaa;\n",
       "    --link-hover: #1a466c;\n",
       "}\n",
       "\n",
       ":root[theme=\"dark\"], :root [data-jp-theme-light=\"false\"], .dataframe_dark{\n",
       "    --background: #303030;\n",
       "    --background-odd: #3c3c3c;\n",
       "    --background-hover: #464646;\n",
       "    --header-text-color: #dddddd;\n",
       "    --text-color: #b3b3b3;\n",
       "    --text-color-dark: #dddddd;\n",
       "    --text-color-medium: #b2b2b2;\n",
       "    --text-color-pale: #737373;\n",
       "    --inner-border-color: #707070;\n",
       "    --bold-border-color: #777777;\n",
       "    --link-color: #008dc0;\n",
       "    --link-color-pale: #97e1fb;\n",
       "    --link-hover: #00688e;\n",
       "}\n",
       "\n",
       "p.dataframe_description {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "table.dataframe {\n",
       "    font-family: \"Helvetica Neue\", Helvetica, Arial, sans-serif;\n",
       "    font-size: 12px;\n",
       "    background-color: var(--background);\n",
       "    color: var(--text-color-dark);\n",
       "    border: none;\n",
       "    border-collapse: collapse;\n",
       "}\n",
       "\n",
       "table.dataframe th, td {\n",
       "    padding: 6px;\n",
       "    border: 1px solid transparent;\n",
       "    text-align: left;\n",
       "}\n",
       "\n",
       "table.dataframe th {\n",
       "    background-color: var(--background);\n",
       "    color: var(--header-text-color);\n",
       "}\n",
       "\n",
       "table.dataframe td {\n",
       "    vertical-align: top;\n",
       "}\n",
       "\n",
       "table.dataframe th.bottomBorder {\n",
       "    border-bottom-color: var(--bold-border-color);\n",
       "}\n",
       "\n",
       "table.dataframe tbody > tr:nth-child(odd) {\n",
       "    background: var(--background-odd);\n",
       "}\n",
       "\n",
       "table.dataframe tbody > tr:nth-child(even) {\n",
       "    background: var(--background);\n",
       "}\n",
       "\n",
       "table.dataframe tbody > tr:hover {\n",
       "    background: var(--background-hover);\n",
       "}\n",
       "\n",
       "table.dataframe a {\n",
       "    cursor: pointer;\n",
       "    color: var(--link-color);\n",
       "    text-decoration: none;\n",
       "}\n",
       "\n",
       "table.dataframe tr:hover > td a {\n",
       "    color: var(--link-color-pale);\n",
       "}\n",
       "\n",
       "table.dataframe a:hover {\n",
       "    color: var(--link-hover);\n",
       "    text-decoration: underline;\n",
       "}\n",
       "\n",
       "table.dataframe img {\n",
       "    max-width: fit-content;\n",
       "}\n",
       "\n",
       "table.dataframe th.complex {\n",
       "    background-color: var(--background);\n",
       "    border: 1px solid var(--background);\n",
       "}\n",
       "\n",
       "table.dataframe .leftBorder {\n",
       "    border-left-color: var(--inner-border-color);\n",
       "}\n",
       "\n",
       "table.dataframe .rightBorder {\n",
       "    border-right-color: var(--inner-border-color);\n",
       "}\n",
       "\n",
       "table.dataframe .rightAlign {\n",
       "    text-align: right;\n",
       "}\n",
       "\n",
       "table.dataframe .expanderSvg {\n",
       "    width: 8px;\n",
       "    height: 8px;\n",
       "    margin-right: 3px;\n",
       "}\n",
       "\n",
       "table.dataframe .expander {\n",
       "    display: flex;\n",
       "    align-items: center;\n",
       "}\n",
       "\n",
       "/* formatting */\n",
       "\n",
       "table.dataframe .null {\n",
       "    color: var(--text-color-pale);\n",
       "}\n",
       "\n",
       "table.dataframe .structural {\n",
       "    color: var(--text-color-medium);\n",
       "    font-weight: bold;\n",
       "}\n",
       "\n",
       "table.dataframe .dataFrameCaption {\n",
       "    font-weight: bold;\n",
       "}\n",
       "\n",
       "table.dataframe .numbers {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "table.dataframe td:hover .formatted .structural, .null {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "table.dataframe tr:hover .formatted .structural, .null {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "\n",
       "\n",
       "\n",
       "            </style>\n",
       "        </head>\n",
       "        <body>\n",
       "            \n",
       "<table class=\"dataframe\" id=\"df_-2046820324\"></table>\n",
       "\n",
       "<p class=\"dataframe_description\">DataFrame: rowsCount = 2, columnsCount = 2</p>\n",
       "        </body>\n",
       "        <script>\n",
       "            \n",
       "/*<!--*/\n",
       "call_DataFrame(function() { DataFrame.addTable({ cols: [{ name: \"<span title=\\\"year: Int\\\">year</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1998</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1999</span></span>\"] }, \n",
       "{ name: \"<span title=\\\"B: Double\\\">B</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">0.000000</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1.569640</span></span>\"] }, \n",
       "{ name: \"<span title=\\\"C: Double\\\">C</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">0.311268</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">0.000000</span></span>\"] }, \n",
       "{ name: \"<span title=\\\"A: Double\\\">A</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">0.000000</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1.129161</span></span>\"] }, \n",
       "{ name: \"<span title=\\\"letter: DataRow\\\">letter</span>\", children: [1, 2, 3], rightAlign: false, values: [\"<span class=\\\"formatted\\\" title=\\\"B: 0.0\\nC: 0.3112683730548147\\nA: 0.0\\\"><span class=\\\"structural\\\">{ </span><span class=\\\"structural\\\">B: </span><span class=\\\"numbers\\\">0.0</span><span class=\\\"structural\\\">, </span><span class=\\\"structural\\\">C: </span><span class=\\\"numbers\\\">0.3</span><span class=\\\"structural\\\">, </span><span class=\\\"structural\\\">A: </span><span class=\\\"numbers\\\">0.0</span><span class=\\\"structural\\\"> }</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"B: 1.5696399986487743\\nC: 0.0\\nA: 1.1291608102568933\\\"><span class=\\\"structural\\\">{ </span><span class=\\\"structural\\\">B: </span><span class=\\\"numbers\\\">1.6</span><span class=\\\"structural\\\">, </span><span class=\\\"structural\\\">C: </span><span class=\\\"numbers\\\">0.0</span><span class=\\\"structural\\\">, </span><span class=\\\"structural\\\">A: </span><span class=\\\"numbers\\\">1.1</span><span class=\\\"structural\\\"> }</span></span>\"] }, \n",
       "], id: -2046820324, rootId: -2046820324, totalRows: 2 } ) });\n",
       "/*-->*/\n",
       "\n",
       "call_DataFrame(function() { DataFrame.renderTable(-2046820324) });\n",
       "\n",
       "\n",
       "        </script>\n",
       "        </html>"
      ]
     },
     "execution_count": 13,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "groupedDf.aggregate { peakToPeak(num.toList()) }"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "f0cc7d51-ffb3-4259-895c-435596bb59d0",
   "metadata": {},
   "source": [
    "And the same thing can be done with pandas `agg` easily with:\n",
    "```python\n",
    "def peak_to_peak(arr):\n",
    "    return arr.max() - arr.min()\n",
    "\n",
    "groupedDf.agg(peak_to_peak)\n",
    "```"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "a4903d09-a8ab-462a-aba5-9fb5ced9e2ed",
   "metadata": {},
   "source": [
    "Kotlin DataFrame is very similar in behavior with pandas, and the transition between the two platform is very natural because of the same strategies implementations."
   ]
  },
  {
   "cell_type": "markdown",
   "id": "63d55b12-25ed-4f30-9868-3020e4c7b83d",
   "metadata": {},
   "source": [
    "### `apply`"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "41ccca1a-6e8b-41ea-82e4-079539f3b239",
   "metadata": {},
   "source": [
    "In python, the `agg` function performs row-wise operations, whereas the `apply` function is a more general purpose. In Kotlin we are more limited with the use of `aggregate`, but with some turnarounds, we can stille get the same results, effortlessly.\n",
    "\n",
    "Consider this python function:\n",
    "```python\n",
    "def top(df, n=5, column):\n",
    "    return df.sort_values(by=column)[-n:]\n",
    "```\n",
    "That returns the top `n` rows with the largest value in `column`.\n",
    "\n",
    "With `apply`, we can call this operation to every group:\n",
    "```python\n",
    "df.groupby(df['year'].dt.year, 'letter').apply(top, n=1, 'num')\n",
    "```\n",
    "\n",
    "and the \"dataframe\" operation is computed along each groups.\n",
    "\n",
    "The same behavior can be accomplished in kotlin with:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 14,
   "id": "33578f1f-3274-484a-a527-5cdd23f366c8",
   "metadata": {
    "tags": []
   },
   "outputs": [],
   "source": [
    "fun top(df: AnyFrame, col: String, n: Int = 5): AnyFrame = df.sortByDesc(col).head(n)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 15,
   "id": "a9f17583-8a1f-41e0-bde5-2ccb681ce202",
   "metadata": {
    "tags": []
   },
   "outputs": [
    {
     "data": {
      "application/kotlindataframe+json": "{\"nrow\":2,\"ncol\":2,\"columns\":[\"year\",\"letter\"],\"kotlin_dataframe\":[{\"year\":1998,\"letter\":{\"B\":[{\"letter\":\"B\",\"num\":0.6246292191371761,\"date\":\"1998-01-02\"}],\"C\":[{\"letter\":\"C\",\"num\":1.4376621993807677,\"date\":\"1998-12-25\"},{\"letter\":\"C\",\"num\":1.126393826325953,\"date\":\"1998-06-22\"}],\"A\":[{\"letter\":\"A\",\"num\":1.9917663922907851,\"date\":\"1998-01-20\"}]}},{\"year\":1999,\"letter\":{\"B\":[{\"letter\":\"B\",\"num\":1.4920787730184157,\"date\":\"1999-12-24\"},{\"letter\":\"B\",\"num\":0.6762208162903859,\"date\":\"1999-03-12\"},{\"letter\":\"B\",\"num\":-0.07756122563035872,\"date\":\"1999-06-24\"}],\"C\":[{\"letter\":\"C\",\"num\":-1.5985374620272976,\"date\":\"1999-02-09\"}],\"A\":[{\"letter\":\"A\",\"num\":0.27096900216864345,\"date\":\"1999-12-20\"},{\"letter\":\"A\",\"num\":-0.8581918080882499,\"date\":\"1999-11-21\"}]}}]}",
      "text/html": [
       "        <html>\n",
       "        <head>\n",
       "            <style type=\"text/css\">\n",
       "                :root {\n",
       "    --background: #fff;\n",
       "    --background-odd: #f5f5f5;\n",
       "    --background-hover: #d9edfd;\n",
       "    --header-text-color: #474747;\n",
       "    --text-color: #848484;\n",
       "    --text-color-dark: #000;\n",
       "    --text-color-medium: #737373;\n",
       "    --text-color-pale: #b3b3b3;\n",
       "    --inner-border-color: #aaa;\n",
       "    --bold-border-color: #000;\n",
       "    --link-color: #296eaa;\n",
       "    --link-color-pale: #296eaa;\n",
       "    --link-hover: #1a466c;\n",
       "}\n",
       "\n",
       ":root[theme=\"dark\"], :root [data-jp-theme-light=\"false\"], .dataframe_dark{\n",
       "    --background: #303030;\n",
       "    --background-odd: #3c3c3c;\n",
       "    --background-hover: #464646;\n",
       "    --header-text-color: #dddddd;\n",
       "    --text-color: #b3b3b3;\n",
       "    --text-color-dark: #dddddd;\n",
       "    --text-color-medium: #b2b2b2;\n",
       "    --text-color-pale: #737373;\n",
       "    --inner-border-color: #707070;\n",
       "    --bold-border-color: #777777;\n",
       "    --link-color: #008dc0;\n",
       "    --link-color-pale: #97e1fb;\n",
       "    --link-hover: #00688e;\n",
       "}\n",
       "\n",
       "p.dataframe_description {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "table.dataframe {\n",
       "    font-family: \"Helvetica Neue\", Helvetica, Arial, sans-serif;\n",
       "    font-size: 12px;\n",
       "    background-color: var(--background);\n",
       "    color: var(--text-color-dark);\n",
       "    border: none;\n",
       "    border-collapse: collapse;\n",
       "}\n",
       "\n",
       "table.dataframe th, td {\n",
       "    padding: 6px;\n",
       "    border: 1px solid transparent;\n",
       "    text-align: left;\n",
       "}\n",
       "\n",
       "table.dataframe th {\n",
       "    background-color: var(--background);\n",
       "    color: var(--header-text-color);\n",
       "}\n",
       "\n",
       "table.dataframe td {\n",
       "    vertical-align: top;\n",
       "}\n",
       "\n",
       "table.dataframe th.bottomBorder {\n",
       "    border-bottom-color: var(--bold-border-color);\n",
       "}\n",
       "\n",
       "table.dataframe tbody > tr:nth-child(odd) {\n",
       "    background: var(--background-odd);\n",
       "}\n",
       "\n",
       "table.dataframe tbody > tr:nth-child(even) {\n",
       "    background: var(--background);\n",
       "}\n",
       "\n",
       "table.dataframe tbody > tr:hover {\n",
       "    background: var(--background-hover);\n",
       "}\n",
       "\n",
       "table.dataframe a {\n",
       "    cursor: pointer;\n",
       "    color: var(--link-color);\n",
       "    text-decoration: none;\n",
       "}\n",
       "\n",
       "table.dataframe tr:hover > td a {\n",
       "    color: var(--link-color-pale);\n",
       "}\n",
       "\n",
       "table.dataframe a:hover {\n",
       "    color: var(--link-hover);\n",
       "    text-decoration: underline;\n",
       "}\n",
       "\n",
       "table.dataframe img {\n",
       "    max-width: fit-content;\n",
       "}\n",
       "\n",
       "table.dataframe th.complex {\n",
       "    background-color: var(--background);\n",
       "    border: 1px solid var(--background);\n",
       "}\n",
       "\n",
       "table.dataframe .leftBorder {\n",
       "    border-left-color: var(--inner-border-color);\n",
       "}\n",
       "\n",
       "table.dataframe .rightBorder {\n",
       "    border-right-color: var(--inner-border-color);\n",
       "}\n",
       "\n",
       "table.dataframe .rightAlign {\n",
       "    text-align: right;\n",
       "}\n",
       "\n",
       "table.dataframe .expanderSvg {\n",
       "    width: 8px;\n",
       "    height: 8px;\n",
       "    margin-right: 3px;\n",
       "}\n",
       "\n",
       "table.dataframe .expander {\n",
       "    display: flex;\n",
       "    align-items: center;\n",
       "}\n",
       "\n",
       "/* formatting */\n",
       "\n",
       "table.dataframe .null {\n",
       "    color: var(--text-color-pale);\n",
       "}\n",
       "\n",
       "table.dataframe .structural {\n",
       "    color: var(--text-color-medium);\n",
       "    font-weight: bold;\n",
       "}\n",
       "\n",
       "table.dataframe .dataFrameCaption {\n",
       "    font-weight: bold;\n",
       "}\n",
       "\n",
       "table.dataframe .numbers {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "table.dataframe td:hover .formatted .structural, .null {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "table.dataframe tr:hover .formatted .structural, .null {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "\n",
       "\n",
       "\n",
       "            </style>\n",
       "        </head>\n",
       "        <body>\n",
       "            \n",
       "<table class=\"dataframe\" id=\"df_-2046820323\"></table>\n",
       "\n",
       "<p class=\"dataframe_description\">DataFrame: rowsCount = 2, columnsCount = 2</p>\n",
       "        </body>\n",
       "        <script>\n",
       "            \n",
       "/*<!--*/\n",
       "call_DataFrame(function() { DataFrame.addTable({ cols: [{ name: \"<span title=\\\"year: Int\\\">year</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1998</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1999</span></span>\"] }, \n",
       "{ name: \"<span title=\\\"B: DataFrame\\\">B</span>\", children: [], rightAlign: false, values: [{ frameId: -2046820322, value: \"<b>DataFrame 1 x 3</b>\" },{ frameId: -2046820321, value: \"<b>DataFrame 3 x 3</b>\" }] }, \n",
       "{ name: \"<span title=\\\"C: DataFrame\\\">C</span>\", children: [], rightAlign: false, values: [{ frameId: -2046820320, value: \"<b>DataFrame 2 x 3</b>\" },{ frameId: -2046820319, value: \"<b>DataFrame 1 x 3</b>\" }] }, \n",
       "{ name: \"<span title=\\\"A: DataFrame\\\">A</span>\", children: [], rightAlign: false, values: [{ frameId: -2046820318, value: \"<b>DataFrame 1 x 3</b>\" },{ frameId: -2046820317, value: \"<b>DataFrame 2 x 3</b>\" }] }, \n",
       "{ name: \"<span title=\\\"letter: DataRow\\\">letter</span>\", children: [1, 2, 3], rightAlign: false, values: [\"<span class=\\\"formatted\\\" title=\\\"B:    letter      num       date\\n 0      B 0.624629 1998-01-02\\n\\nC:    letter      num       date\\n 0      C 1.437662 1998-12-25\\n 1      C 1.126394 1998-06-22\\n\\nA:    letter      num       date\\n 0      A 1.991766 1998-01-20\\n\\\"><span class=\\\"structural\\\">{ </span><span class=\\\"structural\\\">B: </span><span class=\\\"dataFrameCaption\\\">DataFrame [1 x 3]</span><span class=\\\"structural\\\">, </span><span class=\\\"structural\\\">C: </span><span class=\\\"dataFrameCaption\\\">Da<span class=\\\"structural\\\">...</span></span><span class=\\\"structural\\\">, </span><span class=\\\"structural\\\">A...</span><span class=\\\"structural\\\"> }</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"B:    letter       num       date\\n 0      B  1.492079 1999-12-24\\n 1      B  0.676221 1999-03-12\\n 2      B -0.077561 1999-06-24\\n\\nC:    letter       num       date\\n 0      C -1.598537 1999-02-09\\n\\nA:    letter       num       date\\n 0      A  0.270969 1999-12-20\\n 1      A -0.858192 1999-11-21\\n\\\"><span class=\\\"structural\\\">{ </span><span class=\\\"structural\\\">B: </span><span class=\\\"dataFrameCaption\\\">DataFrame [3 x 3]</span><span class=\\\"structural\\\">, </span><span class=\\\"structural\\\">C: </span><span class=\\\"dataFrameCaption\\\">Da<span class=\\\"structural\\\">...</span></span><span class=\\\"structural\\\">, </span><span class=\\\"structural\\\">A...</span><span class=\\\"structural\\\"> }</span></span>\"] }, \n",
       "], id: -2046820323, rootId: -2046820323, totalRows: 2 } ) });\n",
       "/*-->*/\n",
       "\n",
       "/*<!--*/\n",
       "call_DataFrame(function() { DataFrame.addTable({ cols: [{ name: \"<span title=\\\"letter: String\\\">letter</span>\", children: [], rightAlign: false, values: [\"B\"] }, \n",
       "{ name: \"<span title=\\\"num: Double\\\">num</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">0.624629</span></span>\"] }, \n",
       "{ name: \"<span title=\\\"date: kotlinx.datetime.LocalDate\\\">date</span>\", children: [], rightAlign: false, values: [\"1998-01-02\"] }, \n",
       "], id: -2046820322, rootId: -2046820323, totalRows: 1 } ) });\n",
       "/*-->*/\n",
       "\n",
       "/*<!--*/\n",
       "call_DataFrame(function() { DataFrame.addTable({ cols: [{ name: \"<span title=\\\"letter: String\\\">letter</span>\", children: [], rightAlign: false, values: [\"B\",\"B\",\"B\"] }, \n",
       "{ name: \"<span title=\\\"num: Double\\\">num</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1.492079</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">0.676221</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">-0.077561</span></span>\"] }, \n",
       "{ name: \"<span title=\\\"date: kotlinx.datetime.LocalDate\\\">date</span>\", children: [], rightAlign: false, values: [\"1999-12-24\",\"1999-03-12\",\"1999-06-24\"] }, \n",
       "], id: -2046820321, rootId: -2046820323, totalRows: 3 } ) });\n",
       "/*-->*/\n",
       "\n",
       "/*<!--*/\n",
       "call_DataFrame(function() { DataFrame.addTable({ cols: [{ name: \"<span title=\\\"letter: String\\\">letter</span>\", children: [], rightAlign: false, values: [\"C\",\"C\"] }, \n",
       "{ name: \"<span title=\\\"num: Double\\\">num</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1.437662</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1.126394</span></span>\"] }, \n",
       "{ name: \"<span title=\\\"date: kotlinx.datetime.LocalDate\\\">date</span>\", children: [], rightAlign: false, values: [\"1998-12-25\",\"1998-06-22\"] }, \n",
       "], id: -2046820320, rootId: -2046820323, totalRows: 2 } ) });\n",
       "/*-->*/\n",
       "\n",
       "/*<!--*/\n",
       "call_DataFrame(function() { DataFrame.addTable({ cols: [{ name: \"<span title=\\\"letter: String\\\">letter</span>\", children: [], rightAlign: false, values: [\"C\"] }, \n",
       "{ name: \"<span title=\\\"num: Double\\\">num</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">-1.598537</span></span>\"] }, \n",
       "{ name: \"<span title=\\\"date: kotlinx.datetime.LocalDate\\\">date</span>\", children: [], rightAlign: false, values: [\"1999-02-09\"] }, \n",
       "], id: -2046820319, rootId: -2046820323, totalRows: 1 } ) });\n",
       "/*-->*/\n",
       "\n",
       "/*<!--*/\n",
       "call_DataFrame(function() { DataFrame.addTable({ cols: [{ name: \"<span title=\\\"letter: String\\\">letter</span>\", children: [], rightAlign: false, values: [\"A\"] }, \n",
       "{ name: \"<span title=\\\"num: Double\\\">num</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1.991766</span></span>\"] }, \n",
       "{ name: \"<span title=\\\"date: kotlinx.datetime.LocalDate\\\">date</span>\", children: [], rightAlign: false, values: [\"1998-01-20\"] }, \n",
       "], id: -2046820318, rootId: -2046820323, totalRows: 1 } ) });\n",
       "/*-->*/\n",
       "\n",
       "/*<!--*/\n",
       "call_DataFrame(function() { DataFrame.addTable({ cols: [{ name: \"<span title=\\\"letter: String\\\">letter</span>\", children: [], rightAlign: false, values: [\"A\",\"A\"] }, \n",
       "{ name: \"<span title=\\\"num: Double\\\">num</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">0.270969</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">-0.858192</span></span>\"] }, \n",
       "{ name: \"<span title=\\\"date: kotlinx.datetime.LocalDate\\\">date</span>\", children: [], rightAlign: false, values: [\"1999-12-20\",\"1999-11-21\"] }, \n",
       "], id: -2046820317, rootId: -2046820323, totalRows: 2 } ) });\n",
       "/*-->*/\n",
       "\n",
       "call_DataFrame(function() { DataFrame.renderTable(-2046820323) });\n",
       "\n",
       "\n",
       "        </script>\n",
       "        </html>"
      ]
     },
     "execution_count": 15,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "groupedDf.aggregate { top(it, col=\"num\", n=5) }"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "604d3140-882d-4ee4-bf1f-45b85f3de2a7",
   "metadata": {},
   "source": [
    "Where the function invocation is a little more explicit than the panda's one.\n",
    "\n",
    "In contrast to Kotlin DataFrame, pandas `groupby` and `apply` function are way more powerful when grouping and applying functions when slicing data up into buckets with bins or by sample  quantiles (using the function `cut` and `qcut`). Kotlin DataFrame does not provide a way to handle `Categorical` object easily as in pandas, but some turnarounds can be made in order to emulate the behavior of pandas. "
   ]
  },
  {
   "cell_type": "markdown",
   "id": "2e0f0c87-b300-4efa-a2ba-429af2e02e22",
   "metadata": {},
   "source": [
    "## Pivot Tables and Cross Tabulation"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "1ec91b06-3339-4b94-b62d-88f741ca65a5",
   "metadata": {},
   "source": [
    "The pandas method `pivot_table` summarizes the joint use of `groupBy` and `pivot` in Kotlin Dataframe:\n",
    "- The list of values specified in the pivot table, are the ones used for compute an aggregate function in Kotlin DataFrame's `aggregate` function body.\n",
    "- The `index` list provided to `pivot_table`, is represented by the list of column names inside the Kotlin DataFrame `groupBy` method.\n",
    "- The `columns` list is represented by the list of column names inside the Kotlin DataFrame's `pivot` function.\n",
    "- The `aggfunc` parameter in Kotlin is specified inside the `aggregate` function body.\n",
    "\n",
    "Let's consider some examples with the `tips` dataset. (More about this dataset in the examples at the end of this document)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 50,
   "id": "1f268aa8-a047-4c3e-98c6-e6f7d093d65f",
   "metadata": {
    "tags": []
   },
   "outputs": [
    {
     "data": {
      "application/kotlindataframe+json": "{\"nrow\":5,\"ncol\":6,\"columns\":[\"total_bill\",\"tip\",\"smoker\",\"day\",\"time\",\"size\"],\"kotlin_dataframe\":[{\"total_bill\":16.99,\"tip\":1.01,\"smoker\":false,\"day\":\"Sun\",\"time\":\"Dinner\",\"size\":2},{\"total_bill\":10.34,\"tip\":1.66,\"smoker\":false,\"day\":\"Sun\",\"time\":\"Dinner\",\"size\":3},{\"total_bill\":21.01,\"tip\":3.5,\"smoker\":false,\"day\":\"Sun\",\"time\":\"Dinner\",\"size\":3},{\"total_bill\":23.68,\"tip\":3.31,\"smoker\":false,\"day\":\"Sun\",\"time\":\"Dinner\",\"size\":2},{\"total_bill\":24.59,\"tip\":3.61,\"smoker\":false,\"day\":\"Sun\",\"time\":\"Dinner\",\"size\":4}]}",
      "text/html": [
       "        <html>\n",
       "        <head>\n",
       "            <style type=\"text/css\">\n",
       "                :root {\n",
       "    --background: #fff;\n",
       "    --background-odd: #f5f5f5;\n",
       "    --background-hover: #d9edfd;\n",
       "    --header-text-color: #474747;\n",
       "    --text-color: #848484;\n",
       "    --text-color-dark: #000;\n",
       "    --text-color-medium: #737373;\n",
       "    --text-color-pale: #b3b3b3;\n",
       "    --inner-border-color: #aaa;\n",
       "    --bold-border-color: #000;\n",
       "    --link-color: #296eaa;\n",
       "    --link-color-pale: #296eaa;\n",
       "    --link-hover: #1a466c;\n",
       "}\n",
       "\n",
       ":root[theme=\"dark\"], :root [data-jp-theme-light=\"false\"], .dataframe_dark{\n",
       "    --background: #303030;\n",
       "    --background-odd: #3c3c3c;\n",
       "    --background-hover: #464646;\n",
       "    --header-text-color: #dddddd;\n",
       "    --text-color: #b3b3b3;\n",
       "    --text-color-dark: #dddddd;\n",
       "    --text-color-medium: #b2b2b2;\n",
       "    --text-color-pale: #737373;\n",
       "    --inner-border-color: #707070;\n",
       "    --bold-border-color: #777777;\n",
       "    --link-color: #008dc0;\n",
       "    --link-color-pale: #97e1fb;\n",
       "    --link-hover: #00688e;\n",
       "}\n",
       "\n",
       "p.dataframe_description {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "table.dataframe {\n",
       "    font-family: \"Helvetica Neue\", Helvetica, Arial, sans-serif;\n",
       "    font-size: 12px;\n",
       "    background-color: var(--background);\n",
       "    color: var(--text-color-dark);\n",
       "    border: none;\n",
       "    border-collapse: collapse;\n",
       "}\n",
       "\n",
       "table.dataframe th, td {\n",
       "    padding: 6px;\n",
       "    border: 1px solid transparent;\n",
       "    text-align: left;\n",
       "}\n",
       "\n",
       "table.dataframe th {\n",
       "    background-color: var(--background);\n",
       "    color: var(--header-text-color);\n",
       "}\n",
       "\n",
       "table.dataframe td {\n",
       "    vertical-align: top;\n",
       "}\n",
       "\n",
       "table.dataframe th.bottomBorder {\n",
       "    border-bottom-color: var(--bold-border-color);\n",
       "}\n",
       "\n",
       "table.dataframe tbody > tr:nth-child(odd) {\n",
       "    background: var(--background-odd);\n",
       "}\n",
       "\n",
       "table.dataframe tbody > tr:nth-child(even) {\n",
       "    background: var(--background);\n",
       "}\n",
       "\n",
       "table.dataframe tbody > tr:hover {\n",
       "    background: var(--background-hover);\n",
       "}\n",
       "\n",
       "table.dataframe a {\n",
       "    cursor: pointer;\n",
       "    color: var(--link-color);\n",
       "    text-decoration: none;\n",
       "}\n",
       "\n",
       "table.dataframe tr:hover > td a {\n",
       "    color: var(--link-color-pale);\n",
       "}\n",
       "\n",
       "table.dataframe a:hover {\n",
       "    color: var(--link-hover);\n",
       "    text-decoration: underline;\n",
       "}\n",
       "\n",
       "table.dataframe img {\n",
       "    max-width: fit-content;\n",
       "}\n",
       "\n",
       "table.dataframe th.complex {\n",
       "    background-color: var(--background);\n",
       "    border: 1px solid var(--background);\n",
       "}\n",
       "\n",
       "table.dataframe .leftBorder {\n",
       "    border-left-color: var(--inner-border-color);\n",
       "}\n",
       "\n",
       "table.dataframe .rightBorder {\n",
       "    border-right-color: var(--inner-border-color);\n",
       "}\n",
       "\n",
       "table.dataframe .rightAlign {\n",
       "    text-align: right;\n",
       "}\n",
       "\n",
       "table.dataframe .expanderSvg {\n",
       "    width: 8px;\n",
       "    height: 8px;\n",
       "    margin-right: 3px;\n",
       "}\n",
       "\n",
       "table.dataframe .expander {\n",
       "    display: flex;\n",
       "    align-items: center;\n",
       "}\n",
       "\n",
       "/* formatting */\n",
       "\n",
       "table.dataframe .null {\n",
       "    color: var(--text-color-pale);\n",
       "}\n",
       "\n",
       "table.dataframe .structural {\n",
       "    color: var(--text-color-medium);\n",
       "    font-weight: bold;\n",
       "}\n",
       "\n",
       "table.dataframe .dataFrameCaption {\n",
       "    font-weight: bold;\n",
       "}\n",
       "\n",
       "table.dataframe .numbers {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "table.dataframe td:hover .formatted .structural, .null {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "table.dataframe tr:hover .formatted .structural, .null {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "\n",
       "\n",
       "\n",
       "            </style>\n",
       "        </head>\n",
       "        <body>\n",
       "            \n",
       "<table class=\"dataframe\" id=\"df_-2046820264\"></table>\n",
       "\n",
       "<p class=\"dataframe_description\">DataFrame: rowsCount = 5, columnsCount = 6</p>\n",
       "        </body>\n",
       "        <script>\n",
       "            \n",
       "/*<!--*/\n",
       "call_DataFrame(function() { DataFrame.addTable({ cols: [{ name: \"<span title=\\\"total_bill: Double\\\">total_bill</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">16.99</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">10.34</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">21.01</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">23.68</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">24.59</span></span>\"] }, \n",
       "{ name: \"<span title=\\\"tip: Double\\\">tip</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1.01</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1.66</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">3.50</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">3.31</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">3.61</span></span>\"] }, \n",
       "{ name: \"<span title=\\\"smoker: Boolean\\\">smoker</span>\", children: [], rightAlign: false, values: [\"false\",\"false\",\"false\",\"false\",\"false\"] }, \n",
       "{ name: \"<span title=\\\"day: String\\\">day</span>\", children: [], rightAlign: false, values: [\"Sun\",\"Sun\",\"Sun\",\"Sun\",\"Sun\"] }, \n",
       "{ name: \"<span title=\\\"time: String\\\">time</span>\", children: [], rightAlign: false, values: [\"Dinner\",\"Dinner\",\"Dinner\",\"Dinner\",\"Dinner\"] }, \n",
       "{ name: \"<span title=\\\"size: Int\\\">size</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">2</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">3</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">3</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">2</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">4</span></span>\"] }, \n",
       "], id: -2046820264, rootId: -2046820264, totalRows: 5 } ) });\n",
       "/*-->*/\n",
       "\n",
       "call_DataFrame(function() { DataFrame.renderTable(-2046820264) });\n",
       "\n",
       "\n",
       "        </script>\n",
       "        </html>"
      ]
     },
     "execution_count": 50,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "val tips = DataFrame.readCSV(\"../resources/example-datasets/datasets/tips.csv\")\n",
    "tips.head()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 62,
   "id": "1abe17da-2270-415e-bdbf-d8d18f9e53fa",
   "metadata": {
    "tags": []
   },
   "outputs": [
    {
     "data": {
      "application/kotlindataframe+json": "{\"nrow\":8,\"ncol\":5,\"columns\":[\"day\",\"smoker\",\"total_bill\",\"tip\",\"size\"],\"kotlin_dataframe\":[{\"day\":\"Fri\",\"smoker\":false,\"total_bill\":18.419999999999998,\"tip\":2.8125,\"size\":2.25},{\"day\":\"Fri\",\"smoker\":true,\"total_bill\":16.813333333333333,\"tip\":2.714,\"size\":2.066666666666667},{\"day\":\"Sat\",\"smoker\":false,\"total_bill\":19.661777777777786,\"tip\":3.102888888888889,\"size\":2.5555555555555554},{\"day\":\"Sat\",\"smoker\":true,\"total_bill\":21.27666666666666,\"tip\":2.8754761904761903,\"size\":2.4761904761904763},{\"day\":\"Sun\",\"smoker\":false,\"total_bill\":20.506666666666668,\"tip\":3.1678947368421055,\"size\":2.9298245614035086},{\"day\":\"Sun\",\"smoker\":true,\"total_bill\":24.12,\"tip\":3.5168421052631573,\"size\":2.5789473684210527},{\"day\":\"Thur\",\"smoker\":false,\"total_bill\":17.113111111111103,\"tip\":2.673777777777778,\"size\":2.488888888888889},{\"day\":\"Thur\",\"smoker\":true,\"total_bill\":19.19058823529412,\"tip\":3.0299999999999994,\"size\":2.3529411764705883}]}",
      "text/html": [
       "        <html>\n",
       "        <head>\n",
       "            <style type=\"text/css\">\n",
       "                :root {\n",
       "    --background: #fff;\n",
       "    --background-odd: #f5f5f5;\n",
       "    --background-hover: #d9edfd;\n",
       "    --header-text-color: #474747;\n",
       "    --text-color: #848484;\n",
       "    --text-color-dark: #000;\n",
       "    --text-color-medium: #737373;\n",
       "    --text-color-pale: #b3b3b3;\n",
       "    --inner-border-color: #aaa;\n",
       "    --bold-border-color: #000;\n",
       "    --link-color: #296eaa;\n",
       "    --link-color-pale: #296eaa;\n",
       "    --link-hover: #1a466c;\n",
       "}\n",
       "\n",
       ":root[theme=\"dark\"], :root [data-jp-theme-light=\"false\"], .dataframe_dark{\n",
       "    --background: #303030;\n",
       "    --background-odd: #3c3c3c;\n",
       "    --background-hover: #464646;\n",
       "    --header-text-color: #dddddd;\n",
       "    --text-color: #b3b3b3;\n",
       "    --text-color-dark: #dddddd;\n",
       "    --text-color-medium: #b2b2b2;\n",
       "    --text-color-pale: #737373;\n",
       "    --inner-border-color: #707070;\n",
       "    --bold-border-color: #777777;\n",
       "    --link-color: #008dc0;\n",
       "    --link-color-pale: #97e1fb;\n",
       "    --link-hover: #00688e;\n",
       "}\n",
       "\n",
       "p.dataframe_description {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "table.dataframe {\n",
       "    font-family: \"Helvetica Neue\", Helvetica, Arial, sans-serif;\n",
       "    font-size: 12px;\n",
       "    background-color: var(--background);\n",
       "    color: var(--text-color-dark);\n",
       "    border: none;\n",
       "    border-collapse: collapse;\n",
       "}\n",
       "\n",
       "table.dataframe th, td {\n",
       "    padding: 6px;\n",
       "    border: 1px solid transparent;\n",
       "    text-align: left;\n",
       "}\n",
       "\n",
       "table.dataframe th {\n",
       "    background-color: var(--background);\n",
       "    color: var(--header-text-color);\n",
       "}\n",
       "\n",
       "table.dataframe td {\n",
       "    vertical-align: top;\n",
       "}\n",
       "\n",
       "table.dataframe th.bottomBorder {\n",
       "    border-bottom-color: var(--bold-border-color);\n",
       "}\n",
       "\n",
       "table.dataframe tbody > tr:nth-child(odd) {\n",
       "    background: var(--background-odd);\n",
       "}\n",
       "\n",
       "table.dataframe tbody > tr:nth-child(even) {\n",
       "    background: var(--background);\n",
       "}\n",
       "\n",
       "table.dataframe tbody > tr:hover {\n",
       "    background: var(--background-hover);\n",
       "}\n",
       "\n",
       "table.dataframe a {\n",
       "    cursor: pointer;\n",
       "    color: var(--link-color);\n",
       "    text-decoration: none;\n",
       "}\n",
       "\n",
       "table.dataframe tr:hover > td a {\n",
       "    color: var(--link-color-pale);\n",
       "}\n",
       "\n",
       "table.dataframe a:hover {\n",
       "    color: var(--link-hover);\n",
       "    text-decoration: underline;\n",
       "}\n",
       "\n",
       "table.dataframe img {\n",
       "    max-width: fit-content;\n",
       "}\n",
       "\n",
       "table.dataframe th.complex {\n",
       "    background-color: var(--background);\n",
       "    border: 1px solid var(--background);\n",
       "}\n",
       "\n",
       "table.dataframe .leftBorder {\n",
       "    border-left-color: var(--inner-border-color);\n",
       "}\n",
       "\n",
       "table.dataframe .rightBorder {\n",
       "    border-right-color: var(--inner-border-color);\n",
       "}\n",
       "\n",
       "table.dataframe .rightAlign {\n",
       "    text-align: right;\n",
       "}\n",
       "\n",
       "table.dataframe .expanderSvg {\n",
       "    width: 8px;\n",
       "    height: 8px;\n",
       "    margin-right: 3px;\n",
       "}\n",
       "\n",
       "table.dataframe .expander {\n",
       "    display: flex;\n",
       "    align-items: center;\n",
       "}\n",
       "\n",
       "/* formatting */\n",
       "\n",
       "table.dataframe .null {\n",
       "    color: var(--text-color-pale);\n",
       "}\n",
       "\n",
       "table.dataframe .structural {\n",
       "    color: var(--text-color-medium);\n",
       "    font-weight: bold;\n",
       "}\n",
       "\n",
       "table.dataframe .dataFrameCaption {\n",
       "    font-weight: bold;\n",
       "}\n",
       "\n",
       "table.dataframe .numbers {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "table.dataframe td:hover .formatted .structural, .null {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "table.dataframe tr:hover .formatted .structural, .null {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "\n",
       "\n",
       "\n",
       "            </style>\n",
       "        </head>\n",
       "        <body>\n",
       "            \n",
       "<table class=\"dataframe\" id=\"df_-2046820224\"></table>\n",
       "\n",
       "<p class=\"dataframe_description\">DataFrame: rowsCount = 8, columnsCount = 5</p>\n",
       "        </body>\n",
       "        <script>\n",
       "            \n",
       "/*<!--*/\n",
       "call_DataFrame(function() { DataFrame.addTable({ cols: [{ name: \"<span title=\\\"day: String\\\">day</span>\", children: [], rightAlign: false, values: [\"Fri\",\"Fri\",\"Sat\",\"Sat\",\"Sun\",\"Sun\",\"Thur\",\"Thur\"] }, \n",
       "{ name: \"<span title=\\\"smoker: Boolean\\\">smoker</span>\", children: [], rightAlign: false, values: [\"false\",\"true\",\"false\",\"true\",\"false\",\"true\",\"false\",\"true\"] }, \n",
       "{ name: \"<span title=\\\"total_bill: Double\\\">total_bill</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">18.420000</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">16.813333</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">19.661778</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">21.276667</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">20.506667</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">24.120000</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">17.113111</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">19.190588</span></span>\"] }, \n",
       "{ name: \"<span title=\\\"tip: Double\\\">tip</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">2.812500</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">2.714000</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">3.102889</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">2.875476</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">3.167895</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">3.516842</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">2.673778</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">3.030000</span></span>\"] }, \n",
       "{ name: \"<span title=\\\"size: Double\\\">size</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">2.250000</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">2.066667</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">2.555556</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">2.476190</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">2.929825</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">2.578947</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">2.488889</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">2.352941</span></span>\"] }, \n",
       "], id: -2046820224, rootId: -2046820224, totalRows: 8 } ) });\n",
       "/*-->*/\n",
       "\n",
       "call_DataFrame(function() { DataFrame.renderTable(-2046820224) });\n",
       "\n",
       "\n",
       "        </script>\n",
       "        </html>"
      ]
     },
     "execution_count": 62,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "tips.groupBy { day and smoker }.sortBy { day and smoker }.mean()"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "e9b8a8f9-84f8-4e3d-a4e4-b369875e7f6a",
   "metadata": {},
   "source": [
    "In python we could have accomplished the same result with the use of pivot table as follows:\n",
    "```python\n",
    "tips.pivot_table(index=['day', 'smoker'])\n",
    "```\n",
    "and by default, the aggregation function used is the mean of the data. \n",
    "\n",
    "Note that in the following examples we are using the `sortBy` method for sticking to the output that pandas would provide.\n",
    "\n",
    "We could now create a more sophisticated pivot table with this python snippet:\n",
    "```python\n",
    "tips['tips_pct'] = tips['tip'] / tips['total_bill']\n",
    "tips.pivot_table(['tips_pct', 'size'], index=['time', 'day'],\n",
    "                 columns='smoker')\n",
    "```\n",
    "\n",
    "And in Kotlin that pivot table can be created with:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 90,
   "id": "71658f68-cc35-46f3-ad2d-3a251ee6b4ff",
   "metadata": {
    "tags": []
   },
   "outputs": [
    {
     "data": {
      "application/kotlindataframe+json": "{\"nrow\":6,\"ncol\":3,\"columns\":[\"time\",\"day\",\"smoker\"],\"kotlin_dataframe\":[{\"time\":\"Dinner\",\"day\":\"Fri\",\"smoker\":{\"true\":{\"tips_pct\":0.16534735890795094,\"size\":2.2222222222222223},\"false\":{\"tips_pct\":0.13962236590021224,\"size\":2.0}}},{\"time\":\"Dinner\",\"day\":\"Sat\",\"smoker\":{\"true\":{\"tips_pct\":0.14790606514920887,\"size\":2.4761904761904763},\"false\":{\"tips_pct\":0.15804765754377617,\"size\":2.5555555555555554}}},{\"time\":\"Dinner\",\"day\":\"Sun\",\"smoker\":{\"true\":{\"tips_pct\":0.18725031625235425,\"size\":2.5789473684210527},\"false\":{\"tips_pct\":0.16011294305072807,\"size\":2.9298245614035086}}},{\"time\":\"Dinner\",\"day\":\"Thur\",\"smoker\":{\"true\":{\"tips_pct\":null,\"size\":null},\"false\":{\"tips_pct\":0.1597444089456869,\"size\":2.0}}},{\"time\":\"Lunch\",\"day\":\"Fri\",\"smoker\":{\"true\":{\"tips_pct\":0.18893659162141405,\"size\":1.8333333333333333},\"false\":{\"tips_pct\":0.18773466833541927,\"size\":3.0}}},{\"time\":\"Lunch\",\"day\":\"Thur\",\"smoker\":{\"true\":{\"tips_pct\":0.16386327213341778,\"size\":2.3529411764705883},\"false\":{\"tips_pct\":0.16031066524359844,\"size\":2.5}}}]}",
      "text/html": [
       "        <html>\n",
       "        <head>\n",
       "            <style type=\"text/css\">\n",
       "                :root {\n",
       "    --background: #fff;\n",
       "    --background-odd: #f5f5f5;\n",
       "    --background-hover: #d9edfd;\n",
       "    --header-text-color: #474747;\n",
       "    --text-color: #848484;\n",
       "    --text-color-dark: #000;\n",
       "    --text-color-medium: #737373;\n",
       "    --text-color-pale: #b3b3b3;\n",
       "    --inner-border-color: #aaa;\n",
       "    --bold-border-color: #000;\n",
       "    --link-color: #296eaa;\n",
       "    --link-color-pale: #296eaa;\n",
       "    --link-hover: #1a466c;\n",
       "}\n",
       "\n",
       ":root[theme=\"dark\"], :root [data-jp-theme-light=\"false\"], .dataframe_dark{\n",
       "    --background: #303030;\n",
       "    --background-odd: #3c3c3c;\n",
       "    --background-hover: #464646;\n",
       "    --header-text-color: #dddddd;\n",
       "    --text-color: #b3b3b3;\n",
       "    --text-color-dark: #dddddd;\n",
       "    --text-color-medium: #b2b2b2;\n",
       "    --text-color-pale: #737373;\n",
       "    --inner-border-color: #707070;\n",
       "    --bold-border-color: #777777;\n",
       "    --link-color: #008dc0;\n",
       "    --link-color-pale: #97e1fb;\n",
       "    --link-hover: #00688e;\n",
       "}\n",
       "\n",
       "p.dataframe_description {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "table.dataframe {\n",
       "    font-family: \"Helvetica Neue\", Helvetica, Arial, sans-serif;\n",
       "    font-size: 12px;\n",
       "    background-color: var(--background);\n",
       "    color: var(--text-color-dark);\n",
       "    border: none;\n",
       "    border-collapse: collapse;\n",
       "}\n",
       "\n",
       "table.dataframe th, td {\n",
       "    padding: 6px;\n",
       "    border: 1px solid transparent;\n",
       "    text-align: left;\n",
       "}\n",
       "\n",
       "table.dataframe th {\n",
       "    background-color: var(--background);\n",
       "    color: var(--header-text-color);\n",
       "}\n",
       "\n",
       "table.dataframe td {\n",
       "    vertical-align: top;\n",
       "}\n",
       "\n",
       "table.dataframe th.bottomBorder {\n",
       "    border-bottom-color: var(--bold-border-color);\n",
       "}\n",
       "\n",
       "table.dataframe tbody > tr:nth-child(odd) {\n",
       "    background: var(--background-odd);\n",
       "}\n",
       "\n",
       "table.dataframe tbody > tr:nth-child(even) {\n",
       "    background: var(--background);\n",
       "}\n",
       "\n",
       "table.dataframe tbody > tr:hover {\n",
       "    background: var(--background-hover);\n",
       "}\n",
       "\n",
       "table.dataframe a {\n",
       "    cursor: pointer;\n",
       "    color: var(--link-color);\n",
       "    text-decoration: none;\n",
       "}\n",
       "\n",
       "table.dataframe tr:hover > td a {\n",
       "    color: var(--link-color-pale);\n",
       "}\n",
       "\n",
       "table.dataframe a:hover {\n",
       "    color: var(--link-hover);\n",
       "    text-decoration: underline;\n",
       "}\n",
       "\n",
       "table.dataframe img {\n",
       "    max-width: fit-content;\n",
       "}\n",
       "\n",
       "table.dataframe th.complex {\n",
       "    background-color: var(--background);\n",
       "    border: 1px solid var(--background);\n",
       "}\n",
       "\n",
       "table.dataframe .leftBorder {\n",
       "    border-left-color: var(--inner-border-color);\n",
       "}\n",
       "\n",
       "table.dataframe .rightBorder {\n",
       "    border-right-color: var(--inner-border-color);\n",
       "}\n",
       "\n",
       "table.dataframe .rightAlign {\n",
       "    text-align: right;\n",
       "}\n",
       "\n",
       "table.dataframe .expanderSvg {\n",
       "    width: 8px;\n",
       "    height: 8px;\n",
       "    margin-right: 3px;\n",
       "}\n",
       "\n",
       "table.dataframe .expander {\n",
       "    display: flex;\n",
       "    align-items: center;\n",
       "}\n",
       "\n",
       "/* formatting */\n",
       "\n",
       "table.dataframe .null {\n",
       "    color: var(--text-color-pale);\n",
       "}\n",
       "\n",
       "table.dataframe .structural {\n",
       "    color: var(--text-color-medium);\n",
       "    font-weight: bold;\n",
       "}\n",
       "\n",
       "table.dataframe .dataFrameCaption {\n",
       "    font-weight: bold;\n",
       "}\n",
       "\n",
       "table.dataframe .numbers {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "table.dataframe td:hover .formatted .structural, .null {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "table.dataframe tr:hover .formatted .structural, .null {\n",
       "    color: var(--text-color-dark);\n",
       "}\n",
       "\n",
       "\n",
       "\n",
       "\n",
       "            </style>\n",
       "        </head>\n",
       "        <body>\n",
       "            \n",
       "<table class=\"dataframe\" id=\"df_-2046819897\"></table>\n",
       "\n",
       "<p class=\"dataframe_description\">DataFrame: rowsCount = 6, columnsCount = 3</p>\n",
       "        </body>\n",
       "        <script>\n",
       "            \n",
       "/*<!--*/\n",
       "call_DataFrame(function() { DataFrame.addTable({ cols: [{ name: \"<span title=\\\"time: String\\\">time</span>\", children: [], rightAlign: false, values: [\"Dinner\",\"Dinner\",\"Dinner\",\"Dinner\",\"Lunch\",\"Lunch\"] }, \n",
       "{ name: \"<span title=\\\"day: String\\\">day</span>\", children: [], rightAlign: false, values: [\"Fri\",\"Sat\",\"Sun\",\"Thur\",\"Fri\",\"Thur\"] }, \n",
       "{ name: \"<span title=\\\"tips_pct: Double?\\\">tips_pct</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">0.165347</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">0.147906</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">0.187250</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"null\\\">null</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">0.188937</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">0.163863</span></span>\"] }, \n",
       "{ name: \"<span title=\\\"size: Double?\\\">size</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">2.222222</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">2.476190</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">2.578947</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"null\\\">null</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">1.833333</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">2.352941</span></span>\"] }, \n",
       "{ name: \"<span title=\\\"true: DataRow\\\">true</span>\", children: [2, 3], rightAlign: false, values: [\"<span class=\\\"formatted\\\" title=\\\"tips_pct: 0.16534735890795094\\nsize: 2.2222222222222223\\\"><span class=\\\"structural\\\">{ </span><span class=\\\"structural\\\">tips_pct: </span><span class=\\\"numbers\\\">0.2</span><span class=\\\"structural\\\">, </span><span class=\\\"structural\\\">size: </span><span class=\\\"numbers\\\">2.2</span><span class=\\\"structural\\\"> }</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"tips_pct: 0.14790606514920887\\nsize: 2.4761904761904763\\\"><span class=\\\"structural\\\">{ </span><span class=\\\"structural\\\">tips_pct: </span><span class=\\\"numbers\\\">0.1</span><span class=\\\"structural\\\">, </span><span class=\\\"structural\\\">size: </span><span class=\\\"numbers\\\">2.5</span><span class=\\\"structural\\\"> }</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"tips_pct: 0.18725031625235425\\nsize: 2.5789473684210527\\\"><span class=\\\"structural\\\">{ </span><span class=\\\"structural\\\">tips_pct: </span><span class=\\\"numbers\\\">0.2</span><span class=\\\"structural\\\">, </span><span class=\\\"structural\\\">size: </span><span class=\\\"numbers\\\">2.6</span><span class=\\\"structural\\\"> }</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"null\\\">{ }</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"tips_pct: 0.18893659162141405\\nsize: 1.8333333333333333\\\"><span class=\\\"structural\\\">{ </span><span class=\\\"structural\\\">tips_pct: </span><span class=\\\"numbers\\\">0.2</span><span class=\\\"structural\\\">, </span><span class=\\\"structural\\\">size: </span><span class=\\\"numbers\\\">1.8</span><span class=\\\"structural\\\"> }</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"tips_pct: 0.16386327213341778\\nsize: 2.3529411764705883\\\"><span class=\\\"structural\\\">{ </span><span class=\\\"structural\\\">tips_pct: </span><span class=\\\"numbers\\\">0.2</span><span class=\\\"structural\\\">, </span><span class=\\\"structural\\\">size: </span><span class=\\\"numbers\\\">2.4</span><span class=\\\"structural\\\"> }</span></span>\"] }, \n",
       "{ name: \"<span title=\\\"tips_pct: Double\\\">tips_pct</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">0.139622</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">0.158048</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">0.160113</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">0.159744</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">0.187735</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">0.160311</span></span>\"] }, \n",
       "{ name: \"<span title=\\\"size: Double\\\">size</span>\", children: [], rightAlign: true, values: [\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">2.000000</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">2.555556</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">2.929825</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">2.000000</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">3.000000</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"\\\"><span class=\\\"numbers\\\">2.500000</span></span>\"] }, \n",
       "{ name: \"<span title=\\\"false: DataRow\\\">false</span>\", children: [5, 6], rightAlign: false, values: [\"<span class=\\\"formatted\\\" title=\\\"tips_pct: 0.13962236590021224\\nsize: 2.0\\\"><span class=\\\"structural\\\">{ </span><span class=\\\"structural\\\">tips_pct: </span><span class=\\\"numbers\\\">0.1</span><span class=\\\"structural\\\">, </span><span class=\\\"structural\\\">size: </span><span class=\\\"numbers\\\">2.0</span><span class=\\\"structural\\\"> }</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"tips_pct: 0.15804765754377617\\nsize: 2.5555555555555554\\\"><span class=\\\"structural\\\">{ </span><span class=\\\"structural\\\">tips_pct: </span><span class=\\\"numbers\\\">0.2</span><span class=\\\"structural\\\">, </span><span class=\\\"structural\\\">size: </span><span class=\\\"numbers\\\">2.6</span><span class=\\\"structural\\\"> }</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"tips_pct: 0.16011294305072807\\nsize: 2.9298245614035086\\\"><span class=\\\"structural\\\">{ </span><span class=\\\"structural\\\">tips_pct: </span><span class=\\\"numbers\\\">0.2</span><span class=\\\"structural\\\">, </span><span class=\\\"structural\\\">size: </span><span class=\\\"numbers\\\">2.9</span><span class=\\\"structural\\\"> }</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"tips_pct: 0.1597444089456869\\nsize: 2.0\\\"><span class=\\\"structural\\\">{ </span><span class=\\\"structural\\\">tips_pct: </span><span class=\\\"numbers\\\">0.2</span><span class=\\\"structural\\\">, </span><span class=\\\"structural\\\">size: </span><span class=\\\"numbers\\\">2.0</span><span class=\\\"structural\\\"> }</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"tips_pct: 0.18773466833541927\\nsize: 3.0\\\"><span class=\\\"structural\\\">{ </span><span class=\\\"structural\\\">tips_pct: </span><span class=\\\"numbers\\\">0.2</span><span class=\\\"structural\\\">, </span><span class=\\\"structural\\\">size: </span><span class=\\\"numbers\\\">3.0</span><span class=\\\"structural\\\"> }</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"tips_pct: 0.16031066524359844\\nsize: 2.5\\\"><span class=\\\"structural\\\">{ </span><span class=\\\"structural\\\">tips_pct: </span><span class=\\\"numbers\\\">0.2</span><span class=\\\"structural\\\">, </span><span class=\\\"structural\\\">size: </span><span class=\\\"numbers\\\">2.5</span><span class=\\\"structural\\\"> }</span></span>\"] }, \n",
       "{ name: \"<span title=\\\"smoker: DataRow\\\">smoker</span>\", children: [4, 7], rightAlign: false, values: [\"<span class=\\\"formatted\\\" title=\\\"true: { tips_pct:0.165347, size:2.222222 }\\nfalse: { tips_pct:0.139622, size:2.000000 }\\\"><span class=\\\"structural\\\">{ </span><span class=\\\"structural\\\">true: </span><span class=\\\"structural\\\">{ </span><span class=\\\"structural\\\">tips_pct: </span><span class=\\\"numbers\\\">0.2</span><span class=\\\"structural\\\">, </span><span class=\\\"structural\\\">si...</span><span class=\\\"structural\\\"> }</span><span class=\\\"structural\\\">, </span><span class=\\\"structural\\\">f...</span><span class=\\\"structural\\\"> }</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"true: { tips_pct:0.147906, size:2.476190 }\\nfalse: { tips_pct:0.158048, size:2.555556 }\\\"><span class=\\\"structural\\\">{ </span><span class=\\\"structural\\\">true: </span><span class=\\\"structural\\\">{ </span><span class=\\\"structural\\\">tips_pct: </span><span class=\\\"numbers\\\">0.1</span><span class=\\\"structural\\\">, </span><span class=\\\"structural\\\">si...</span><span class=\\\"structural\\\"> }</span><span class=\\\"structural\\\">, </span><span class=\\\"structural\\\">f...</span><span class=\\\"structural\\\"> }</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"true: { tips_pct:0.187250, size:2.578947 }\\nfalse: { tips_pct:0.160113, size:2.929825 }\\\"><span class=\\\"structural\\\">{ </span><span class=\\\"structural\\\">true: </span><span class=\\\"structural\\\">{ </span><span class=\\\"structural\\\">tips_pct: </span><span class=\\\"numbers\\\">0.2</span><span class=\\\"structural\\\">, </span><span class=\\\"structural\\\">si...</span><span class=\\\"structural\\\"> }</span><span class=\\\"structural\\\">, </span><span class=\\\"structural\\\">f...</span><span class=\\\"structural\\\"> }</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"false: { tips_pct:0.159744, size:2.000000 }\\\"><span class=\\\"structural\\\">{ </span><span class=\\\"structural\\\">false: </span><span class=\\\"structural\\\">{ </span><span class=\\\"structural\\\">tips_pct: </span><span class=\\\"numbers\\\">0.2</span><span class=\\\"structural\\\">, </span><span class=\\\"structural\\\">size: </span><span class=\\\"numbers\\\">2.0</span><span class=\\\"structural\\\"> }</span><span class=\\\"structural\\\"> }</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"true: { tips_pct:0.188937, size:1.833333 }\\nfalse: { tips_pct:0.187735, size:3.000000 }\\\"><span class=\\\"structural\\\">{ </span><span class=\\\"structural\\\">true: </span><span class=\\\"structural\\\">{ </span><span class=\\\"structural\\\">tips_pct: </span><span class=\\\"numbers\\\">0.2</span><span class=\\\"structural\\\">, </span><span class=\\\"structural\\\">si...</span><span class=\\\"structural\\\"> }</span><span class=\\\"structural\\\">, </span><span class=\\\"structural\\\">f...</span><span class=\\\"structural\\\"> }</span></span>\",\"<span class=\\\"formatted\\\" title=\\\"true: { tips_pct:0.163863, size:2.352941 }\\nfalse: { tips_pct:0.160311, size:2.500000 }\\\"><span class=\\\"structural\\\">{ </span><span class=\\\"structural\\\">true: </span><span class=\\\"structural\\\">{ </span><span class=\\\"structural\\\">tips_pct: </span><span class=\\\"numbers\\\">0.2</span><span class=\\\"structural\\\">, </span><span class=\\\"structural\\\">si...</span><span class=\\\"structural\\\"> }</span><span class=\\\"structural\\\">, </span><span class=\\\"structural\\\">f...</span><span class=\\\"structural\\\"> }</span></span>\"] }, \n",
       "], id: -2046819897, rootId: -2046819897, totalRows: 6 } ) });\n",
       "/*-->*/\n",
       "\n",
       "call_DataFrame(function() { DataFrame.renderTable(-2046819897) });\n",
       "\n",
       "\n",
       "        </script>\n",
       "        </html>"
      ]
     },
     "execution_count": 90,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "tips.add(\"tips_pct\") { tip / total_bill }\n",
    "    .groupBy { time and day }.sortBy { time and day }\n",
    "    .pivot { smoker }\n",
    "    .aggregate {\n",
    "        mean(\"tips_pct\") into \"tips_pct\" \n",
    "        mean { size } into \"size\"\n",
    "    }"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "e8da40f8-feec-4798-8181-29259ea7398e",
   "metadata": {},
   "source": [
    "So a deep understanding of the pandas `pivot_table` method, can really helps a lot when it comes to learning how to pivot and rearrange the dataset using Kotlin DataFrame!"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "e693b72b-6d73-4f89-bdd0-4afff133f2c8",
   "metadata": {},
   "source": [
    "## Conclusion"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "e82300b8-248c-46e0-96e2-32a192aa4d9d",
   "metadata": {},
   "source": [
    "This process of data aggregation and the application of the split-apply-combine strategy can help both data cleaning as well as modeling or statistical analysis work, and as we have seen in this chapter, Kotlin DataFrame offers almost the same pandas capabilities when it come to those operations, especially when pivoting tables. "
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Kotlin",
   "language": "kotlin",
   "name": "kotlin"
  },
  "language_info": {
   "codemirror_mode": "text/x-kotlin",
   "file_extension": ".kt",
   "mimetype": "text/x-kotlin",
   "name": "kotlin",
   "nbconvert_exporter": "",
   "pygments_lexer": "kotlin",
   "version": "1.8.20"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 5
}
